/* Pango
 * pangopdflib-fontmap.c:
 *
 * Copyright (C) 2000 Red Hat Software
 * Copyright (C) 2000 Tor Lillqvist
 * Copyright (C) 2001 Sun Microsystems
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "config.h"

#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#ifdef HAVE_DIRENT_H
#include <dirent.h>
#endif

#include <fontconfig/fontconfig.h>

#include "pango-fontmap.h"
#include "pango-utils.h"
#include "pangopdflib-private.h"
#include "pangopdflib-context.h"
#include "modules.h"

#ifdef G_OS_WIN32
#define STRICT
#include <windows.h>
#endif

typedef struct _PangoPDFLIBFamily       PangoPDFLIBFamily;

struct _PangoPDFLIBFontMap
{
  PangoFontMap parent_instance;

  PangoPDFLIBFontEmbed font_embed;

  FT_Library library;

  /* We have one map from  PangoFontDescription -> PangoXftPatternSet
   * per language tag.
   */
  GList *fontset_hash_list;
  /* pattern_hash is used to make sure we only store one copy of
   * each identical pattern. (Speeds up lookup).
   */
  GHashTable *pattern_hash; 
  GHashTable *coverage_hash; /* Maps font file name -> PangoCoverage */

  GHashTable *fonts; /* Maps XftPattern -> PangoPDFLIBFont */
	
  GQueue *fontset_cache;	/* Recently used fontsets */

  /* List of all families availible */
  PangoPDFLIBFamily **families;
  int n_families;		/* -1 == uninitialized */

  double dpi_x;
  double dpi_y;

  guint closed : 1;

  /* Function to call on prepared patterns to do final
   * config tweaking.
   */
  PangoPDFLIBSubstituteFunc substitute_func;
  gpointer substitute_data;
  GDestroyNotify substitute_destroy;
};

/************************************************************
 *              Code shared with PangoXft                   *
 ************************************************************/

#define PangoFcFamily      PangoPDFLIBFamily
#define _PangoFcFamily      _PangoPDFLIBFamily
#define PangoFcFontMap     PangoPDFLIBFontMap
#define PangoFcFont     PangoPDFLIBFont

#define PANGO_FC_FONT_MAP PANGO_PDFLIB_FONT_MAP

#define pango_fc_font_map_get_type pango_pdflib_font_map_get_type
#define _pango_fc_font_map_add _pango_pdflib_font_map_add
#define _pango_fc_font_map_remove _pango_pdflib_font_map_remove
#define _pango_fc_font_map_get_coverage _pango_pdflib_font_map_get_coverage
#define _pango_fc_font_map_set_coverage _pango_pdflib_font_map_set_coverage
#define _pango_fc_font_desc_from_pattern _pango_pdflib_font_desc_from_pattern
#define _pango_fc_font_new               _pango_pdflib_font_new

#define PANGO_FC_NAME "PangoPDFLIB"

#include "pangofc-fontmap.cI"

/*************************************************************
 *                  FreeType specific code                   *
 *************************************************************/

static PangoPDFLIBFontMap *pango_pdflib_global_fontmap = NULL;

/**
 * pango_pdflib_font_map_new:
 * 
 * Create a new #PangoPDFLIBFontMap object; a fontmap is used
 * to cache information about available fonts, and holds
 * certain global parameters such as the resolution and
 * the default substitute function (see
 * pango_font_map_set_default_substitute ()).
 * 
 * Return value: the newly created fontmap object. Unref
 * with g_object_unref when you are finished with it.
 **/
PangoFontMap *
pango_pdflib_font_map_new (void)
{
  static gboolean registered_modules = FALSE;
  PangoPDFLIBFontMap *pdflibfontmap;
  FT_Error error;
  
  if (!registered_modules)
    {
      int i;
      
      registered_modules = TRUE;
      
      /* Make sure that the type system is initialized */
      g_type_init ();
  
      for (i = 0; _pango_included_pdflib_modules[i].list; i++)
        pango_module_register (&_pango_included_pdflib_modules[i]);
    }

  pdflibfontmap = g_object_new (PANGO_TYPE_PDFLIB_FONT_MAP, NULL);
  
  error = FT_Init_FreeType (&pdflibfontmap->library);
  if (error != FT_Err_Ok)
    {
      g_warning ("Error from FT_Init_FreeType: %s",
		 _pango_pdflib_ft_strerror (error));
      return NULL;
    }

  return (PangoFontMap *)pdflibfontmap;
}

/**
 * pango_pdflib_font_map_set_default_substitute:
 * @fontmap: a #PangoPDFLIBFontMap
 * @func: function to call to to do final config tweaking
 *        on #FcPattern objects.
 * @data: data to pass to @func
 * @notify: function to call when @data is no longer used.
 * 
 * Sets a function that will be called to do final configuration
 * substitution on a #FcPattern before it is used to load
 * the font. This function can be used to do things like set
 * hinting and antialiasing options.
 *
 * Since: 1.2
 **/
void
pango_pdflib_font_map_set_default_substitute (PangoPDFLIBFontMap        *fontmap,
					      PangoPDFLIBSubstituteFunc  func,
					      gpointer                   data,
					      GDestroyNotify             notify)
{
  if (fontmap->substitute_destroy)
    fontmap->substitute_destroy (fontmap->substitute_data);

  fontmap->substitute_func = func;
  fontmap->substitute_data = data;
  fontmap->substitute_destroy = notify;
  
  pango_fc_font_map_cache_clear (fontmap);
}

/**
 * pango_pdflib_font_map_substitute_changed:
 * @fontmap: a #PangoPDFLIBFontmap
 * 
 * Call this function any time the results of the
 * default substitution function set with
 * pango_pdflib_font_map_set_default_substitute() change.
 * That is, if your subsitution function will return different
 * results for the same input pattern, you must call this function.
 **/
void
pango_pdflib_font_map_substitute_changed (PangoPDFLIBFontMap *fontmap)
{
  pango_fc_font_map_cache_clear (fontmap);
}

PangoPDFLIBFontEmbed
pango_pdflib_font_map_get_font_embed (PangoPDFLIBFontMap *font_map)
{
  g_return_val_if_fail (PANGO_PDFLIB_IS_FONT_MAP (font_map),
			PANGO_PDFLIB_FONT_EMBED_INVALID);

  return font_map->font_embed;
}

void
pango_pdflib_font_map_set_font_embed (PangoPDFLIBFontMap  *font_map,
				      PangoPDFLIBFontEmbed font_embed)
{
  g_return_if_fail (PANGO_PDFLIB_IS_FONT_MAP (font_map));

  font_map->font_embed = font_embed;
}

/**
 * pango_pdflib_font_map_set_resolution:
 * @fontmap: a #PangoPDFLIBFontmap 
 * @dpi_x: dots per inch in the X direction
 * @dpi_y: dots per inch in the Y direction
 * 
 * Sets the horizontal and vertical resolutions for the fontmap.
 **/
void
pango_pdflib_font_map_set_resolution (PangoPDFLIBFontMap *fontmap,
				      double           dpi_x,
				      double           dpi_y)
{
  g_return_if_fail (PANGO_PDFLIB_IS_FONT_MAP (fontmap));
  
  fontmap->dpi_x = dpi_x;
  fontmap->dpi_y = dpi_y;

  pango_pdflib_font_map_substitute_changed (fontmap);
}

/**
 * pango_pdflib_font_map_create_context_with_debug_flags:
 * @fontmap:     a #PangoPDFLIBFontmap.
 * @debug_flags: #PangoPDFLIBDebugFlags.
 * 
 * Create a #PangoContext for the given fontmap.
 * 
 * Debug flags are currently ignored.
 *
 * Return value: the newly created context; free with g_object_unref().
 **/
PangoContext *
pango_pdflib_font_map_create_context_with_debug_flags (PangoPDFLIBFontMap *fontmap,
						       PangoPDFLIBDebugFlags debug_flags)
{
  PangoContext *context;

  g_return_val_if_fail (PANGO_PDFLIB_IS_FONT_MAP (fontmap), NULL);
  
  context = pango_pdflib_context_new ();
  pango_context_set_font_map (PANGO_CONTEXT (context),
			      PANGO_FONT_MAP (fontmap));

  return PANGO_CONTEXT (context);
}

/**
 * pango_pdflib_font_map_create_context:
 * @fontmap:    a #PangoPDFLIBFontmap.
 * 
 * Create a #PangoContext for the given fontmap.
 * 
 * Return value: the newly created context; free with g_object_unref().
 **/
PangoContext *
pango_pdflib_font_map_create_context (PangoPDFLIBFontMap *fontmap)
{
  g_return_val_if_fail (PANGO_PDFLIB_IS_FONT_MAP (fontmap), NULL);
  
  return pango_pdflib_font_map_create_context_with_debug_flags (fontmap,
								PANGO_PDFLIB_DEBUG_NONE);
}

/**
 * pango_pdflib_get_context:
 * @dpi_x:  the horizontal dpi of the target device
 * @dpi_y:  the vertical dpi of the target device
 * 
 * Retrieves a #PangoContext for the default PangoPDFLIB fontmap
 * (see pango_pdflib_fontmap_get_for_display()) and sets the resolution
 * for the default fontmap to @dpi_x by @dpi_y.
 *
 * Use of this function is discouraged, see pango_pdflib_font_map_create_context()
 * instead.
 * 
 * Return value: the new #PangoContext
 **/
PangoContext *
pango_pdflib_get_context (double                   dpi_x,
			  double                   dpi_y)
{
  PangoFontMap *fontmap;
  
  fontmap = pango_pdflib_font_map_for_display ();
  pango_pdflib_font_map_set_resolution (PANGO_PDFLIB_FONT_MAP (fontmap), dpi_x, dpi_y);

  return pango_pdflib_font_map_create_context (PANGO_PDFLIB_FONT_MAP (fontmap));
}

/**
 * pango_pdflib_font_map_for_display:
 *
 * Returns a #PangoPDFLIBFontMap. This font map is cached and should
 * not be freed. If the font map is no longer needed, it can
 * be released with pango_pdflib_shutdown_display().
 *
 * Returns: a #PangoPDFLIBFontMap.
 **/
PangoFontMap *
pango_pdflib_font_map_for_display (void)
{
  if (pango_pdflib_global_fontmap != NULL)
    return PANGO_FONT_MAP (pango_pdflib_global_fontmap);

  pango_pdflib_global_fontmap = (PangoPDFLIBFontMap *)pango_pdflib_font_map_new ();

  return PANGO_FONT_MAP (pango_pdflib_global_fontmap);
}

/**
 * pango_pdflib_shutdown_display:
 * 
 * Free the global fontmap. (See pango_pdflib_font_map_for_display())
 **/
void
pango_pdflib_shutdown_display (void)
{
  if (pango_pdflib_global_fontmap)
    {
      pango_fc_font_map_cache_clear (pango_pdflib_global_fontmap);
      
      g_object_unref (G_OBJECT (pango_pdflib_global_fontmap));
      
      pango_pdflib_global_fontmap = NULL;
    }
}

FT_Library
_pango_pdflib_font_map_get_library (PangoFontMap *fontmap)
{
  PangoPDFLIBFontMap *pdflibfontmap = (PangoPDFLIBFontMap *)fontmap;
  
  return pdflibfontmap->library;
}

static void
pango_fc_do_finalize (PangoPDFLIBFontMap    *fontmap)
{
  FT_Done_FreeType (fontmap->library);
}

static void
pango_fc_default_substitute (PangoPDFLIBFontMap *fontmap,
			     FcPattern       *pattern)
{
  FcValue v;
  
  FcConfigSubstitute (NULL, pattern, FcMatchPattern);

  if (fontmap->substitute_func)
    fontmap->substitute_func (pattern, fontmap->substitute_data);

  FcDefaultSubstitute (pattern);
  if (FcPatternGet (pattern, FC_DPI, 0, &v) == FcResultNoMatch)
    FcPatternAddDouble (pattern, FC_DPI, fontmap->dpi_y);
}
