/* Pango
 * pangopdflib.c: Routines for handling FreeType2 fonts with PDFlib library
 *
 * Copyright (C) 1999 Red Hat Software
 * Copyright (C) 2000 Tor Lillqvist
 * Copyright (C) 2001, 2002 Sun Microsystems
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "config.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <glib.h>

#include <ft2build.h>
#include FT_FREETYPE_H

#include "pango-utils.h"
#include "pangopdflib.h"
#include "pangopdflib-private.h"
#include "pango-attributes-private.h"
#include "pango-layout-private.h"
#include "pango-context-private.h"

/* for compatibility with older freetype versions */
#ifndef FT_LOAD_TARGET_MONO
#define FT_LOAD_TARGET_MONO  FT_LOAD_MONOCHROME
#endif

#define PANGO_TYPE_PDFLIB_FONT              (pango_pdflib_font_get_type ())
#define PANGO_PDFLIB_FONT(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), PANGO_TYPE_PDFLIB_FONT, PangoPDFLIBFont))
#define PANGO_PDFLIB_FONT_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), PANGO_TYPE_PDFLIB_FONT, PangoPDFLIBFontClass))
#define PANGO_PDFLIB_IS_FONT(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), PANGO_TYPE_PDFLIB_FONT))
#define PANGO_PDFLIB_IS_FONT_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), PANGO_TYPE_PDFLIB_FONT))
#define PANGO_PDFLIB_FONT_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), PANGO_TYPE_PDFLIB_FONT, PangoPDFLIBFontClass))

typedef struct _PangoPDFLIBFontClass   PangoPDFLIBFontClass;
typedef struct _PangoPDFLIBMetricsInfo PangoPDFLIBMetricsInfo;

struct _PangoPDFLIBFontClass
{
  PangoFontClass parent_class;
};

struct _PangoPDFLIBMetricsInfo
{
  const char       *sample_str;
  PangoFontMetrics *metrics;
};

static PangoFontClass *parent_class;	/* Parent class structure for PangoPDFLIBFont */

static void pango_pdflib_font_class_init (PangoPDFLIBFontClass *class);
static void pango_pdflib_font_init       (PangoPDFLIBFont      *pdflibfont);
static void pango_pdflib_font_finalize   (GObject         *object);

static PangoFontDescription *pango_pdflib_font_describe          (PangoFont            *font);

static PangoEngineShape *    pango_pdflib_font_find_shaper       (PangoFont            *font,
								  PangoLanguage        *language,
								  guint32               ch);

static void                  pango_pdflib_font_get_glyph_extents (PangoFont            *font,
								  PangoGlyph            glyph,
								  PangoRectangle       *ink_rect,
								  PangoRectangle       *logical_rect);

static PangoFontMetrics *    pango_pdflib_font_get_metrics       (PangoFont            *font,
								  PangoLanguage        *language);
  
static GType pango_pdflib_font_get_type (void);

typedef enum {
  DEFAULT,
  AGL1,
  AGL2,
  AGL3,
  AGL4
} PangoPDFLIBEncoding;

typedef struct _GlyphEntry {
  const char *name;
  PangoPDFLIBEncoding encoding;
  guchar code_point;
  gunichar2 unicode;
} GlyphEntry;

static GlyphEntry glyph_entries[] = {
  { "A",		DEFAULT,	0x41,	0x0041 },
  { "AE",		DEFAULT,	0xC6,	0x00C6 },
  { "AEacute",		AGL1,		0x82,	0x01FC },
  { "AEsmall",		AGL4,		0x24,	0xF7E6 },
  { "Aacute",		DEFAULT,	0xC1,	0x00C1 },
  { "Aacutesmall",	AGL4,		0x1F,	0xF7E1 },
  { "Abreve",		AGL1,		0x03,	0x0102 },
  { "Acircumflex",	DEFAULT,	0xC2,	0x00C2 },
  { "Acircumflexsmall",	AGL4,		0x20,	0xF7E2 },
  { "Acute",		AGL3,		0xB6,	0xF6C9 },
  { "Acutesmall",	AGL4,		0x1B,	0xF7B4 },
  { "Adieresis",	DEFAULT,	0xC4,	0x00C4 },
  { "Adieresissmall",	AGL4,		0x22,	0xF7E4 },
  { "Agrave",		DEFAULT,	0xC0,	0x00C0 },
  { "Agravesmall",	AGL4,		0x1E,	0xF7E0 },
  { "Alpha",		AGL1,		0xA3,	0x0391 },
  { "Alphatonos",	AGL1,		0x9A,	0x0386 },
  { "Amacron",		AGL1,		0x01,	0x0100 },
  { "Aogonek",		AGL1,		0x05,	0x0104 },
  { "Aring",		DEFAULT,	0xC5,	0x00C5 },
  { "Aringacute",	AGL1,		0x80,	0x01FA },
  { "Aringsmall",	AGL4,		0x23,	0xF7E5 },
  { "Asmall",		AGL3,		0xFC,	0xF761 },
  { "Atilde",		DEFAULT,	0xC3,	0x00C3 },
  { "Atildesmall",	AGL4,		0x21,	0xF7E3 },
  { "B",		DEFAULT,	0x42,	0x0042 },
  { "Beta",		AGL1,		0xA4,	0x0392 },
  { "Brevesmall",	AGL3,		0xE1,	0xF6F4 },
  { "Bsmall",		AGL3,		0xFD,	0xF762 },
  { "C",		DEFAULT,	0x43,	0x0043 },
  { "Cacute",		AGL1,		0x07,	0x0106 },
  { "Caron",		AGL3,		0xB7,	0xF6CA },
  { "Caronsmall",	AGL3,		0xE2,	0xF6F5 },
  { "Ccaron",		AGL1,		0x0D,	0x010C },
  { "Ccedilla",		DEFAULT,	0xC7,	0x00C7 },
  { "Ccedillasmall",	AGL4,		0x25,	0xF7E7 },
  { "Ccircumflex",	AGL1,		0x09,	0x0108 },
  { "Cdotaccent",	AGL1,		0x0B,	0x010A },
  { "Cedillasmall",	AGL4,		0x1C,	0xF7B8 },
  { "Chi",		AGL1,		0xB8,	0x03A7 },
  { "Circumflexsmall",	AGL3,		0xE3,	0xF6F6 },
  { "Csmall",		AGL3,		0xFE,	0xF763 },
  { "D",		DEFAULT,	0x44,	0x0044 },
  { "Dcaron",		AGL1,		0x0F,	0x010E },
  { "Dcroat",		AGL1,		0x11,	0x0110 },
  { "Delta",		AGL1,		0xA6,	0x0394 },
  { "Delta",		AGL3,		0x35,	0x2206 },
  { "Dieresis",		AGL3,		0xB8,	0xF6CB },
  { "DieresisAcute",	AGL3,		0xB9,	0xF6CC },
  { "DieresisGrave",	AGL3,		0xBA,	0xF6CD },
  { "Dieresissmall",	AGL4,		0x19,	0xF7A8 },
  { "Dotaccentsmall",	AGL3,		0xE4,	0xF6F7 },
  { "Dsmall",		AGL3,		0xFF,	0xF764 },
  { "E",		DEFAULT,	0x45,	0x0045 },
  { "Eacute",		DEFAULT,	0xC9,	0x00C9 },
  { "Eacutesmall",	AGL4,		0x27,	0xF7E9 },
  { "Ebreve",		AGL1,		0x15,	0x0114 },
  { "Ecaron",		AGL1,		0x1B,	0x011A },
  { "Ecircumflex",	DEFAULT,	0xCA,	0x00CA },
  { "Ecircumflexsmall",	AGL4,		0x28,	0xF7EA },
  { "Edieresis",	DEFAULT,	0xCB,	0x00CB },
  { "Edieresissmall",	AGL4,		0x29,	0xF7EB },
  { "Edotaccent",	AGL1,		0x17,	0x0116 },
  { "Egrave",		DEFAULT,	0xC8,	0x00C8 },
  { "Egravesmall",	AGL4,		0x26,	0xF7E8 },
  { "Emacron",		AGL1,		0x13,	0x0112 },
  { "Eng",		AGL1,		0x4B,	0x014A },
  { "Eogonek",		AGL1,		0x19,	0x0118 },
  { "Epsilon",		AGL1,		0xA7,	0x0395 },
  { "Epsilontonos",	AGL1,		0x9C,	0x0388 },
  { "Esmall",		AGL4,		0x01,	0xF765 },
  { "Eta",		AGL1,		0xA9,	0x0397 },
  { "Etatonos",		AGL1,		0x9D,	0x0389 },
  { "Eth",		DEFAULT,	0xD0,	0x00D0 },
  { "Ethsmall",		AGL4,		0x2E,	0xF7F0 },
  { "Euro",		DEFAULT,	0x80,	0x20AC },
  { "F",		DEFAULT,	0x46,	0x0046 },
  { "Fsmall",		AGL4,		0x02,	0xF766 },
  { "G",		DEFAULT,	0x47,	0x0047 },
  { "Gamma",		AGL1,		0xA5,	0x0393 },
  { "Gbreve",		AGL1,		0x1F,	0x011E },
  { "Gcaron",		AGL1,		0x7E,	0x01E6 },
  { "Gcircumflex",	AGL1,		0x1D,	0x011C },
  { "Gcommaaccent",	AGL1,		0x23,	0x0122 },
  { "Gdotaccent",	AGL1,		0x21,	0x0120 },
  { "Grave",		AGL3,		0xBB,	0xF6CE },
  { "Gravesmall",	AGL3,		0xFB,	0xF760 },
  { "Gsmall",		AGL4,		0x03,	0xF767 },
  { "H",		DEFAULT,	0x48,	0x0048 },
  { "H18533",		AGL3,		0x9C,	0x25CF },
  { "H18543",		AGL3,		0x93,	0x25AA },
  { "H18551",		AGL3,		0x94,	0x25AB },
  { "H22073",		AGL3,		0x92,	0x25A1 },
  { "Hbar",		AGL1,		0x27,	0x0126 },
  { "Hcircumflex",	AGL1,		0x25,	0x0124 },
  { "Hsmall",		AGL4,		0x04,	0xF768 },
  { "Hungarumlaut",	AGL3,		0xBC,	0xF6CF },
  { "Hungarumlautsmall",	AGL3,		0xE5,	0xF6F8 },
  { "I",		DEFAULT,	0x49,	0x0049 },
  { "IJ",		AGL1,		0x33,	0x0132 },
  { "Iacute",		DEFAULT,	0xCD,	0x00CD },
  { "Iacutesmall",	AGL4,		0x2B,	0xF7ED },
  { "Ibreve",		AGL1,		0x2D,	0x012C },
  { "Icircumflex",	DEFAULT,	0xCE,	0x00CE },
  { "Icircumflexsmall",	AGL4,		0x2C,	0xF7EE },
  { "Idieresis",	DEFAULT,	0xCF,	0x00CF },
  { "Idieresissmall",	AGL4,		0x2D,	0xF7EF },
  { "Idotaccent",	AGL1,		0x31,	0x0130 },
  { "Ifraktur",		AGL3,		0x15,	0x2111 },
  { "Igrave",		DEFAULT,	0xCC,	0x00CC },
  { "Igravesmall",	AGL4,		0x2A,	0xF7EC },
  { "Imacron",		AGL1,		0x2B,	0x012A },
  { "Iogonek",		AGL1,		0x2F,	0x012E },
  { "Iota",		AGL1,		0xAB,	0x0399 },
  { "Iotadieresis",	AGL1,		0xBB,	0x03AA },
  { "Iotatonos",	AGL1,		0x9E,	0x038A },
  { "Ismall",		AGL4,		0x05,	0xF769 },
  { "Itilde",		AGL1,		0x29,	0x0128 },
  { "J",		DEFAULT,	0x4A,	0x004A },
  { "Jcircumflex",	AGL1,		0x35,	0x0134 },
  { "Jsmall",		AGL4,		0x06,	0xF76A },
  { "K",		DEFAULT,	0x4B,	0x004B },
  { "Kappa",		AGL1,		0xAC,	0x039A },
  { "Kcommaaccent",	AGL1,		0x37,	0x0136 },
  { "Ksmall",		AGL4,		0x07,	0xF76B },
  { "L",		DEFAULT,	0x4C,	0x004C },
  { "LL",		AGL3,		0xAC,	0xF6BF },
  { "Lacute",		AGL1,		0x3A,	0x0139 },
  { "Lambda",		AGL1,		0xAD,	0x039B },
  { "Lcaron",		AGL1,		0x3E,	0x013D },
  { "Lcommaaccent",	AGL1,		0x3C,	0x013B },
  { "Ldot",		AGL1,		0x40,	0x013F },
  { "Lslash",		AGL1,		0x42,	0x0141 },
  { "Lslashsmall",	AGL3,		0xE6,	0xF6F9 },
  { "Lsmall",		AGL4,		0x08,	0xF76C },
  { "M",		DEFAULT,	0x4D,	0x004D },
  { "Macron",		AGL3,		0xBD,	0xF6D0 },
  { "Macronsmall",	AGL4,		0x1A,	0xF7AF },
  { "Msmall",		AGL4,		0x09,	0xF76D },
  { "Mu",		AGL1,		0xAE,	0x039C },
  { "N",		DEFAULT,	0x4E,	0x004E },
  { "Nacute",		AGL1,		0x44,	0x0143 },
  { "Ncaron",		AGL1,		0x48,	0x0147 },
  { "Ncommaaccent",	AGL1,		0x46,	0x0145 },
  { "Nsmall",		AGL4,		0x0A,	0xF76E },
  { "Ntilde",		DEFAULT,	0xD1,	0x00D1 },
  { "Ntildesmall",	AGL4,		0x2F,	0xF7F1 },
  { "Nu",		AGL1,		0xAF,	0x039D },
  { "O",		DEFAULT,	0x4F,	0x004F },
  { "OE",		DEFAULT,	0x8C,	0x0152 },
  { "OEsmall",		AGL3,		0xE7,	0xF6FA },
  { "Oacute",		DEFAULT,	0xD3,	0x00D3 },
  { "Oacutesmall",	AGL4,		0x31,	0xF7F3 },
  { "Obreve",		AGL1,		0x4F,	0x014E },
  { "Ocircumflex",	DEFAULT,	0xD4,	0x00D4 },
  { "Ocircumflexsmall",	AGL4,		0x32,	0xF7F4 },
  { "Odieresis",	DEFAULT,	0xD6,	0x00D6 },
  { "Odieresissmall",	AGL4,		0x34,	0xF7F6 },
  { "Ogoneksmall",	AGL3,		0xE8,	0xF6FB },
  { "Ograve",		DEFAULT,	0xD2,	0x00D2 },
  { "Ogravesmall",	AGL4,		0x30,	0xF7F2 },
  { "Ohorn",		AGL1,		0x7A,	0x01A0 },
  { "Ohungarumlaut",	AGL1,		0x51,	0x0150 },
  { "Omacron",		AGL1,		0x4D,	0x014C },
  { "Omega",		AGL1,		0xBA,	0x03A9 },
  { "Omega",		AGL3,		0x1B,	0x2126 },
  { "Omegatonos",	AGL1,		0xA1,	0x038F },
  { "Omicron",		AGL1,		0xB1,	0x039F },
  { "Omicrontonos",	AGL1,		0x9F,	0x038C },
  { "Oslash",		DEFAULT,	0xD8,	0x00D8 },
  { "Oslashacute",	AGL1,		0x84,	0x01FE },
  { "Oslashsmall",	AGL4,		0x35,	0xF7F8 },
  { "Osmall",		AGL4,		0x0B,	0xF76F },
  { "Otilde",		DEFAULT,	0xD5,	0x00D5 },
  { "Otildesmall",	AGL4,		0x33,	0xF7F5 },
  { "P",		DEFAULT,	0x50,	0x0050 },
  { "Phi",		AGL1,		0xB7,	0x03A6 },
  { "Pi",		AGL1,		0xB2,	0x03A0 },
  { "Psi",		AGL1,		0xB9,	0x03A8 },
  { "Psmall",		AGL4,		0x0C,	0xF770 },
  { "Q",		DEFAULT,	0x51,	0x0051 },
  { "Qsmall",		AGL4,		0x0D,	0xF771 },
  { "R",		DEFAULT,	0x52,	0x0052 },
  { "Racute",		AGL1,		0x53,	0x0154 },
  { "Rcaron",		AGL1,		0x57,	0x0158 },
  { "Rcommaaccent",	AGL1,		0x55,	0x0156 },
  { "Rfraktur",		AGL3,		0x19,	0x211C },
  { "Rho",		AGL1,		0xB3,	0x03A1 },
  { "Ringsmall",	AGL3,		0xE9,	0xF6FC },
  { "Rsmall",		AGL4,		0x0E,	0xF772 },
  { "S",		DEFAULT,	0x53,	0x0053 },
  { "SF010000",		AGL3,		0x63,	0x250C },
  { "SF020000",		AGL3,		0x65,	0x2514 },
  { "SF030000",		AGL3,		0x64,	0x2510 },
  { "SF040000",		AGL3,		0x66,	0x2518 },
  { "SF050000",		AGL3,		0x6B,	0x253C },
  { "SF060000",		AGL3,		0x69,	0x252C },
  { "SF070000",		AGL3,		0x6A,	0x2534 },
  { "SF080000",		AGL3,		0x67,	0x251C },
  { "SF090000",		AGL3,		0x68,	0x2524 },
  { "SF100000",		AGL3,		0x61,	0x2500 },
  { "SF110000",		AGL3,		0x62,	0x2502 },
  { "SF190000",		AGL3,		0x7D,	0x2561 },
  { "SF200000",		AGL3,		0x7E,	0x2562 },
  { "SF210000",		AGL3,		0x72,	0x2556 },
  { "SF220000",		AGL3,		0x71,	0x2555 },
  { "SF230000",		AGL3,		0x7F,	0x2563 },
  { "SF240000",		AGL3,		0x6D,	0x2551 },
  { "SF250000",		AGL3,		0x73,	0x2557 },
  { "SF260000",		AGL3,		0x79,	0x255D },
  { "SF270000",		AGL3,		0x78,	0x255C },
  { "SF280000",		AGL3,		0x77,	0x255B },
  { "SF360000",		AGL3,		0x7A,	0x255E },
  { "SF370000",		AGL3,		0x7B,	0x255F },
  { "SF380000",		AGL3,		0x76,	0x255A },
  { "SF390000",		AGL3,		0x70,	0x2554 },
  { "SF400000",		AGL3,		0x85,	0x2569 },
  { "SF410000",		AGL3,		0x82,	0x2566 },
  { "SF420000",		AGL3,		0x7C,	0x2560 },
  { "SF430000",		AGL3,		0x6C,	0x2550 },
  { "SF440000",		AGL3,		0x88,	0x256C },
  { "SF450000",		AGL3,		0x83,	0x2567 },
  { "SF460000",		AGL3,		0x84,	0x2568 },
  { "SF470000",		AGL3,		0x80,	0x2564 },
  { "SF480000",		AGL3,		0x81,	0x2565 },
  { "SF490000",		AGL3,		0x75,	0x2559 },
  { "SF500000",		AGL3,		0x74,	0x2558 },
  { "SF510000",		AGL3,		0x6E,	0x2552 },
  { "SF520000",		AGL3,		0x6F,	0x2553 },
  { "SF530000",		AGL3,		0x87,	0x256B },
  { "SF540000",		AGL3,		0x86,	0x256A },
  { "Sacute",		AGL1,		0x59,	0x015A },
  { "Scaron",		DEFAULT,	0x8A,	0x0160 },
  { "Scaronsmall",	AGL3,		0xEA,	0xF6FD },
  { "Scedilla",		AGL1,		0x5D,	0x015E },
  { "Scircumflex",	AGL1,		0x5B,	0x015C },
  { "Scommaaccent",	AGL1,		0x86,	0x0218 },
  { "Sigma",		AGL1,		0xB4,	0x03A3 },
  { "Ssmall",		AGL4,		0x0F,	0xF773 },
  { "T",		DEFAULT,	0x54,	0x0054 },
  { "Tau",		AGL1,		0xB5,	0x03A4 },
  { "Tbar",		AGL1,		0x63,	0x0166 },
  { "Tcaron",		AGL1,		0x61,	0x0164 },
  { "Tcommaaccent",	AGL1,		0x5F,	0x0162 },
  { "Theta",		AGL1,		0xAA,	0x0398 },
  { "Thorn",		DEFAULT,	0xDE,	0x00DE },
  { "Thornsmall",	AGL4,		0x3B,	0xF7FE },
  { "Tildesmall",	AGL3,		0xEB,	0xF6FE },
  { "Tsmall",		AGL4,		0x10,	0xF774 },
  { "U",		DEFAULT,	0x55,	0x0055 },
  { "Uacute",		DEFAULT,	0xDA,	0x00DA },
  { "Uacutesmall",	AGL4,		0x37,	0xF7FA },
  { "Ubreve",		AGL1,		0x69,	0x016C },
  { "Ucircumflex",	DEFAULT,	0xDB,	0x00DB },
  { "Ucircumflexsmall",	AGL4,		0x38,	0xF7FB },
  { "Udieresis",	DEFAULT,	0xDC,	0x00DC },
  { "Udieresissmall",	AGL4,		0x39,	0xF7FC },
  { "Ugrave",		DEFAULT,	0xD9,	0x00D9 },
  { "Ugravesmall",	AGL4,		0x36,	0xF7F9 },
  { "Uhorn",		AGL1,		0x7C,	0x01AF },
  { "Uhungarumlaut",	AGL1,		0x6D,	0x0170 },
  { "Umacron",		AGL1,		0x67,	0x016A },
  { "Uogonek",		AGL1,		0x6F,	0x0172 },
  { "Upsilon",		AGL1,		0xB6,	0x03A5 },
  { "Upsilon1",		AGL1,		0xE1,	0x03D2 },
  { "Upsilondieresis",	AGL1,		0xBC,	0x03AB },
  { "Upsilontonos",	AGL1,		0xA0,	0x038E },
  { "Uring",		AGL1,		0x6B,	0x016E },
  { "Usmall",		AGL4,		0x11,	0xF775 },
  { "Utilde",		AGL1,		0x65,	0x0168 },
  { "V",		DEFAULT,	0x56,	0x0056 },
  { "Vsmall",		AGL4,		0x12,	0xF776 },
  { "W",		DEFAULT,	0x57,	0x0057 },
  { "Wacute",		AGL2,		0xE0,	0x1E82 },
  { "Wcircumflex",	AGL1,		0x71,	0x0174 },
  { "Wdieresis",	AGL2,		0xE2,	0x1E84 },
  { "Wgrave",		AGL2,		0xDE,	0x1E80 },
  { "Wsmall",		AGL4,		0x13,	0xF777 },
  { "X",		DEFAULT,	0x58,	0x0058 },
  { "Xi",		AGL1,		0xB0,	0x039E },
  { "Xsmall",		AGL4,		0x14,	0xF778 },
  { "Y",		DEFAULT,	0x59,	0x0059 },
  { "Yacute",		DEFAULT,	0xDD,	0x00DD },
  { "Yacutesmall",	AGL4,		0x3A,	0xF7FD },
  { "Ycircumflex",	AGL1,		0x73,	0x0176 },
  { "Ydieresis",	DEFAULT,	0x9F,	0x0178 },
  { "Ydieresissmall",	AGL4,		0x3C,	0xF7FF },
  { "Ygrave",		AGL2,		0xE4,	0x1EF2 },
  { "Ysmall",		AGL4,		0x15,	0xF779 },
  { "Z",		DEFAULT,	0x5A,	0x005A },
  { "Zacute",		AGL1,		0x75,	0x0179 },
  { "Zcaron",		DEFAULT,	0x8E,	0x017D },
  { "Zcaronsmall",	AGL3,		0xEC,	0xF6FF },
  { "Zdotaccent",	AGL1,		0x77,	0x017B },
  { "Zeta",		AGL1,		0xA8,	0x0396 },
  { "Zsmall",		AGL4,		0x16,	0xF77A },
  { "a",		DEFAULT,	0x61,	0x0061 },
  { "aacute",		DEFAULT,	0xE1,	0x00E1 },
  { "abreve",		AGL1,		0x04,	0x0103 },
  { "acircumflex",	DEFAULT,	0xE2,	0x00E2 },
  { "acute",		DEFAULT,	0xB4,	0x00B4 },
  { "acutecomb",	AGL1,		0x94,	0x0301 },
  { "adieresis",	DEFAULT,	0xE4,	0x00E4 },
  { "ae",		DEFAULT,	0xE6,	0x00E6 },
  { "aeacute",		AGL1,		0x83,	0x01FD },
  { "afii00208",	AGL2,		0xEB,	0x2015 },
  { "afii10017",	AGL2,		0x0F,	0x0410 },
  { "afii10018",	AGL2,		0x10,	0x0411 },
  { "afii10019",	AGL2,		0x11,	0x0412 },
  { "afii10020",	AGL2,		0x12,	0x0413 },
  { "afii10021",	AGL2,		0x13,	0x0414 },
  { "afii10022",	AGL2,		0x14,	0x0415 },
  { "afii10023",	AGL2,		0x01,	0x0401 },
  { "afii10024",	AGL2,		0x15,	0x0416 },
  { "afii10025",	AGL2,		0x16,	0x0417 },
  { "afii10026",	AGL2,		0x17,	0x0418 },
  { "afii10027",	AGL2,		0x18,	0x0419 },
  { "afii10028",	AGL2,		0x19,	0x041A },
  { "afii10029",	AGL2,		0x1A,	0x041B },
  { "afii10030",	AGL2,		0x1B,	0x041C },
  { "afii10031",	AGL2,		0x1C,	0x041D },
  { "afii10032",	AGL2,		0x1D,	0x041E },
  { "afii10033",	AGL2,		0x1E,	0x041F },
  { "afii10034",	AGL2,		0x1F,	0x0420 },
  { "afii10035",	AGL2,		0x20,	0x0421 },
  { "afii10036",	AGL2,		0x21,	0x0422 },
  { "afii10037",	AGL2,		0x22,	0x0423 },
  { "afii10038",	AGL2,		0x23,	0x0424 },
  { "afii10039",	AGL2,		0x24,	0x0425 },
  { "afii10040",	AGL2,		0x25,	0x0426 },
  { "afii10041",	AGL2,		0x26,	0x0427 },
  { "afii10042",	AGL2,		0x27,	0x0428 },
  { "afii10043",	AGL2,		0x28,	0x0429 },
  { "afii10044",	AGL2,		0x29,	0x042A },
  { "afii10045",	AGL2,		0x2A,	0x042B },
  { "afii10046",	AGL2,		0x2B,	0x042C },
  { "afii10047",	AGL2,		0x2C,	0x042D },
  { "afii10048",	AGL2,		0x2D,	0x042E },
  { "afii10049",	AGL2,		0x2E,	0x042F },
  { "afii10050",	AGL2,		0x63,	0x0490 },
  { "afii10051",	AGL2,		0x02,	0x0402 },
  { "afii10052",	AGL2,		0x03,	0x0403 },
  { "afii10053",	AGL2,		0x04,	0x0404 },
  { "afii10054",	AGL2,		0x05,	0x0405 },
  { "afii10055",	AGL2,		0x06,	0x0406 },
  { "afii10056",	AGL2,		0x07,	0x0407 },
  { "afii10057",	AGL2,		0x08,	0x0408 },
  { "afii10058",	AGL2,		0x09,	0x0409 },
  { "afii10059",	AGL2,		0x0A,	0x040A },
  { "afii10060",	AGL2,		0x0B,	0x040B },
  { "afii10061",	AGL2,		0x0C,	0x040C },
  { "afii10062",	AGL2,		0x0D,	0x040E },
  { "afii10063",	AGL3,		0xB1,	0xF6C4 },
  { "afii10064",	AGL3,		0xB2,	0xF6C5 },
  { "afii10065",	AGL2,		0x2F,	0x0430 },
  { "afii10066",	AGL2,		0x30,	0x0431 },
  { "afii10067",	AGL2,		0x31,	0x0432 },
  { "afii10068",	AGL2,		0x32,	0x0433 },
  { "afii10069",	AGL2,		0x33,	0x0434 },
  { "afii10070",	AGL2,		0x34,	0x0435 },
  { "afii10071",	AGL2,		0x4F,	0x0451 },
  { "afii10072",	AGL2,		0x35,	0x0436 },
  { "afii10073",	AGL2,		0x36,	0x0437 },
  { "afii10074",	AGL2,		0x37,	0x0438 },
  { "afii10075",	AGL2,		0x38,	0x0439 },
  { "afii10076",	AGL2,		0x39,	0x043A },
  { "afii10077",	AGL2,		0x3A,	0x043B },
  { "afii10078",	AGL2,		0x3B,	0x043C },
  { "afii10079",	AGL2,		0x3C,	0x043D },
  { "afii10080",	AGL2,		0x3D,	0x043E },
  { "afii10081",	AGL2,		0x3E,	0x043F },
  { "afii10082",	AGL2,		0x3F,	0x0440 },
  { "afii10083",	AGL2,		0x40,	0x0441 },
  { "afii10084",	AGL2,		0x41,	0x0442 },
  { "afii10085",	AGL2,		0x42,	0x0443 },
  { "afii10086",	AGL2,		0x43,	0x0444 },
  { "afii10087",	AGL2,		0x44,	0x0445 },
  { "afii10088",	AGL2,		0x45,	0x0446 },
  { "afii10089",	AGL2,		0x46,	0x0447 },
  { "afii10090",	AGL2,		0x47,	0x0448 },
  { "afii10091",	AGL2,		0x48,	0x0449 },
  { "afii10092",	AGL2,		0x49,	0x044A },
  { "afii10093",	AGL2,		0x4A,	0x044B },
  { "afii10094",	AGL2,		0x4B,	0x044C },
  { "afii10095",	AGL2,		0x4C,	0x044D },
  { "afii10096",	AGL2,		0x4D,	0x044E },
  { "afii10097",	AGL2,		0x4E,	0x044F },
  { "afii10098",	AGL2,		0x64,	0x0491 },
  { "afii10099",	AGL2,		0x50,	0x0452 },
  { "afii10100",	AGL2,		0x51,	0x0453 },
  { "afii10101",	AGL2,		0x52,	0x0454 },
  { "afii10102",	AGL2,		0x53,	0x0455 },
  { "afii10103",	AGL2,		0x54,	0x0456 },
  { "afii10104",	AGL2,		0x55,	0x0457 },
  { "afii10105",	AGL2,		0x56,	0x0458 },
  { "afii10106",	AGL2,		0x57,	0x0459 },
  { "afii10107",	AGL2,		0x58,	0x045A },
  { "afii10108",	AGL2,		0x59,	0x045B },
  { "afii10109",	AGL2,		0x5A,	0x045C },
  { "afii10110",	AGL2,		0x5B,	0x045E },
  { "afii10145",	AGL2,		0x0E,	0x040F },
  { "afii10146",	AGL2,		0x5D,	0x0462 },
  { "afii10147",	AGL2,		0x5F,	0x0472 },
  { "afii10148",	AGL2,		0x61,	0x0474 },
  { "afii10192",	AGL3,		0xB3,	0xF6C6 },
  { "afii10193",	AGL2,		0x5C,	0x045F },
  { "afii10194",	AGL2,		0x5E,	0x0463 },
  { "afii10195",	AGL2,		0x60,	0x0473 },
  { "afii10196",	AGL2,		0x62,	0x0475 },
  { "afii10831",	AGL3,		0xB4,	0xF6C7 },
  { "afii10832",	AGL3,		0xB5,	0xF6C8 },
  { "afii10846",	AGL2,		0x65,	0x04D9 },
  { "afii299",		AGL2,		0xE8,	0x200E },
  { "afii300",		AGL2,		0xE9,	0x200F },
  { "afii301",		AGL2,		0xE7,	0x200D },
  { "afii57381",	AGL2,		0xD1,	0x066A },
  { "afii57388",	AGL2,		0x97,	0x060C },
  { "afii57392",	AGL2,		0xC7,	0x0660 },
  { "afii57393",	AGL2,		0xC8,	0x0661 },
  { "afii57394",	AGL2,		0xC9,	0x0662 },
  { "afii57395",	AGL2,		0xCA,	0x0663 },
  { "afii57396",	AGL2,		0xCB,	0x0664 },
  { "afii57397",	AGL2,		0xCC,	0x0665 },
  { "afii57398",	AGL2,		0xCD,	0x0666 },
  { "afii57399",	AGL2,		0xCE,	0x0667 },
  { "afii57400",	AGL2,		0xCF,	0x0668 },
  { "afii57401",	AGL2,		0xD0,	0x0669 },
  { "afii57403",	AGL2,		0x98,	0x061B },
  { "afii57407",	AGL2,		0x99,	0x061F },
  { "afii57409",	AGL2,		0x9A,	0x0621 },
  { "afii57410",	AGL2,		0x9B,	0x0622 },
  { "afii57411",	AGL2,		0x9C,	0x0623 },
  { "afii57412",	AGL2,		0x9D,	0x0624 },
  { "afii57413",	AGL2,		0x9E,	0x0625 },
  { "afii57414",	AGL2,		0x9F,	0x0626 },
  { "afii57415",	AGL2,		0xA0,	0x0627 },
  { "afii57416",	AGL2,		0xA1,	0x0628 },
  { "afii57417",	AGL2,		0xA2,	0x0629 },
  { "afii57418",	AGL2,		0xA3,	0x062A },
  { "afii57419",	AGL2,		0xA4,	0x062B },
  { "afii57420",	AGL2,		0xA5,	0x062C },
  { "afii57421",	AGL2,		0xA6,	0x062D },
  { "afii57422",	AGL2,		0xA7,	0x062E },
  { "afii57423",	AGL2,		0xA8,	0x062F },
  { "afii57424",	AGL2,		0xA9,	0x0630 },
  { "afii57425",	AGL2,		0xAA,	0x0631 },
  { "afii57426",	AGL2,		0xAB,	0x0632 },
  { "afii57427",	AGL2,		0xAC,	0x0633 },
  { "afii57428",	AGL2,		0xAD,	0x0634 },
  { "afii57429",	AGL2,		0xAE,	0x0635 },
  { "afii57430",	AGL2,		0xAF,	0x0636 },
  { "afii57431",	AGL2,		0xB0,	0x0637 },
  { "afii57432",	AGL2,		0xB1,	0x0638 },
  { "afii57433",	AGL2,		0xB2,	0x0639 },
  { "afii57434",	AGL2,		0xB3,	0x063A },
  { "afii57440",	AGL2,		0xB4,	0x0640 },
  { "afii57441",	AGL2,		0xB5,	0x0641 },
  { "afii57442",	AGL2,		0xB6,	0x0642 },
  { "afii57443",	AGL2,		0xB7,	0x0643 },
  { "afii57444",	AGL2,		0xB8,	0x0644 },
  { "afii57445",	AGL2,		0xB9,	0x0645 },
  { "afii57446",	AGL2,		0xBA,	0x0646 },
  { "afii57448",	AGL2,		0xBC,	0x0648 },
  { "afii57449",	AGL2,		0xBD,	0x0649 },
  { "afii57450",	AGL2,		0xBE,	0x064A },
  { "afii57451",	AGL2,		0xBF,	0x064B },
  { "afii57452",	AGL2,		0xC0,	0x064C },
  { "afii57453",	AGL2,		0xC1,	0x064D },
  { "afii57454",	AGL2,		0xC2,	0x064E },
  { "afii57455",	AGL2,		0xC3,	0x064F },
  { "afii57456",	AGL2,		0xC4,	0x0650 },
  { "afii57457",	AGL2,		0xC5,	0x0651 },
  { "afii57458",	AGL2,		0xC6,	0x0652 },
  { "afii57470",	AGL2,		0xBB,	0x0647 },
  { "afii57505",	AGL2,		0xD9,	0x06A4 },
  { "afii57506",	AGL2,		0xD4,	0x067E },
  { "afii57507",	AGL2,		0xD5,	0x0686 },
  { "afii57508",	AGL2,		0xD8,	0x0698 },
  { "afii57509",	AGL2,		0xDA,	0x06AF },
  { "afii57511",	AGL2,		0xD3,	0x0679 },
  { "afii57512",	AGL2,		0xD6,	0x0688 },
  { "afii57513",	AGL2,		0xD7,	0x0691 },
  { "afii57514",	AGL2,		0xDB,	0x06BA },
  { "afii57519",	AGL2,		0xDC,	0x06D2 },
  { "afii57534",	AGL2,		0xDD,	0x06D5 },
  { "afii57636",	AGL3,		0x12,	0x20AA },
  { "afii57645",	AGL2,		0x73,	0x05BE },
  { "afii57658",	AGL2,		0x78,	0x05C3 },
  { "afii57664",	AGL2,		0x79,	0x05D0 },
  { "afii57665",	AGL2,		0x7A,	0x05D1 },
  { "afii57666",	AGL2,		0x7B,	0x05D2 },
  { "afii57667",	AGL2,		0x7C,	0x05D3 },
  { "afii57668",	AGL2,		0x7D,	0x05D4 },
  { "afii57669",	AGL2,		0x7E,	0x05D5 },
  { "afii57670",	AGL2,		0x7F,	0x05D6 },
  { "afii57671",	AGL2,		0x80,	0x05D7 },
  { "afii57672",	AGL2,		0x81,	0x05D8 },
  { "afii57673",	AGL2,		0x82,	0x05D9 },
  { "afii57674",	AGL2,		0x83,	0x05DA },
  { "afii57675",	AGL2,		0x84,	0x05DB },
  { "afii57676",	AGL2,		0x85,	0x05DC },
  { "afii57677",	AGL2,		0x86,	0x05DD },
  { "afii57678",	AGL2,		0x87,	0x05DE },
  { "afii57679",	AGL2,		0x88,	0x05DF },
  { "afii57680",	AGL2,		0x89,	0x05E0 },
  { "afii57681",	AGL2,		0x8A,	0x05E1 },
  { "afii57682",	AGL2,		0x8B,	0x05E2 },
  { "afii57683",	AGL2,		0x8C,	0x05E3 },
  { "afii57684",	AGL2,		0x8D,	0x05E4 },
  { "afii57685",	AGL2,		0x8E,	0x05E5 },
  { "afii57686",	AGL2,		0x8F,	0x05E6 },
  { "afii57687",	AGL2,		0x90,	0x05E7 },
  { "afii57688",	AGL2,		0x91,	0x05E8 },
  { "afii57689",	AGL2,		0x92,	0x05E9 },
  { "afii57690",	AGL2,		0x93,	0x05EA },
  { "afii57694",	AGL4,		0x5D,	0xFB2A },
  { "afii57695",	AGL4,		0x5E,	0xFB2B },
  { "afii57700",	AGL4,		0x60,	0xFB4B },
  { "afii57705",	AGL4,		0x5C,	0xFB1F },
  { "afii57716",	AGL2,		0x94,	0x05F0 },
  { "afii57717",	AGL2,		0x95,	0x05F1 },
  { "afii57718",	AGL2,		0x96,	0x05F2 },
  { "afii57723",	AGL4,		0x5F,	0xFB35 },
  { "afii57793",	AGL2,		0x6A,	0x05B4 },
  { "afii57794",	AGL2,		0x6B,	0x05B5 },
  { "afii57795",	AGL2,		0x6C,	0x05B6 },
  { "afii57796",	AGL2,		0x70,	0x05BB },
  { "afii57797",	AGL2,		0x6E,	0x05B8 },
  { "afii57798",	AGL2,		0x6D,	0x05B7 },
  { "afii57799",	AGL2,		0x66,	0x05B0 },
  { "afii57800",	AGL2,		0x68,	0x05B2 },
  { "afii57801",	AGL2,		0x67,	0x05B1 },
  { "afii57802",	AGL2,		0x69,	0x05B3 },
  { "afii57803",	AGL2,		0x77,	0x05C2 },
  { "afii57804",	AGL2,		0x76,	0x05C1 },
  { "afii57806",	AGL2,		0x6F,	0x05B9 },
  { "afii57807",	AGL2,		0x71,	0x05BC },
  { "afii57839",	AGL2,		0x72,	0x05BD },
  { "afii57841",	AGL2,		0x74,	0x05BF },
  { "afii57842",	AGL2,		0x75,	0x05C0 },
  { "afii57929",	AGL1,		0x8A,	0x02BC },
  { "afii61248",	AGL3,		0x14,	0x2105 },
  { "afii61289",	AGL3,		0x16,	0x2113 },
  { "afii61352",	AGL3,		0x17,	0x2116 },
  { "afii61573",	AGL2,		0xF0,	0x202C },
  { "afii61574",	AGL2,		0xF1,	0x202D },
  { "afii61575",	AGL2,		0xF2,	0x202E },
  { "afii61664",	AGL2,		0xE6,	0x200C },
  { "afii63167",	AGL2,		0xD2,	0x066D },
  { "afii64937",	AGL1,		0x8B,	0x02BD },
  { "agrave",		DEFAULT,	0xE0,	0x00E0 },
  { "alef",		AGL2,		0x79,	0x05D0 },
  { "aleph",		AGL3,		0x1D,	0x2135 },
  { "alpha",		AGL1,		0xC2,	0x03B1 },
  { "alphatonos",	AGL1,		0xBD,	0x03AC },
  { "amacron",		AGL1,		0x02,	0x0101 },
  { "ampersand",	DEFAULT,	0x26,	0x0026 },
  { "ampersandsmall",	AGL3,		0xEF,	0xF726 },
  { "angle",		AGL3,		0x44,	0x2220 },
  { "angleleft",	AGL3,		0x5F,	0x2329 },
  { "angleright",	AGL3,		0x60,	0x232A },
  { "anoteleia",	AGL1,		0x9B,	0x0387 },
  { "aogonek",		AGL1,		0x06,	0x0105 },
  { "approxequal",	AGL3,		0x4D,	0x2248 },
  { "aring",		DEFAULT,	0xE5,	0x00E5 },
  { "aringacute",	AGL1,		0x81,	0x01FB },
  { "arrowboth",	AGL3,		0x28,	0x2194 },
  { "arrowdblboth",	AGL3,		0x30,	0x21D4 },
  { "arrowdbldown",	AGL3,		0x2F,	0x21D3 },
  { "arrowdblleft",	AGL3,		0x2C,	0x21D0 },
  { "arrowdblright",	AGL3,		0x2E,	0x21D2 },
  { "arrowdblup",	AGL3,		0x2D,	0x21D1 },
  { "arrowdown",	AGL3,		0x27,	0x2193 },
  { "arrowhorizex",	AGL4,		0x3F,	0xF8E7 },
  { "arrowleft",	AGL3,		0x24,	0x2190 },
  { "arrowright",	AGL3,		0x26,	0x2192 },
  { "arrowup",		AGL3,		0x25,	0x2191 },
  { "arrowupdn",	AGL3,		0x29,	0x2195 },
  { "arrowupdnbse",	AGL3,		0x2A,	0x21A8 },
  { "arrowvertex",	AGL4,		0x3E,	0xF8E6 },
  { "asciicircum",	DEFAULT,	0x5E,	0x005E },
  { "asciitilde",	DEFAULT,	0x7E,	0x007E },
  { "asterisk",		DEFAULT,	0x2A,	0x002A },
  { "asteriskmath",	AGL3,		0x3E,	0x2217 },
  { "asuperior",	AGL3,		0xD6,	0xF6E9 },
  { "at",		DEFAULT,	0x40,	0x0040 },
  { "atilde",		DEFAULT,	0xE3,	0x00E3 },
  { "ayin",		AGL2,		0x8B,	0x05E2 },
  { "b",		DEFAULT,	0x62,	0x0062 },
  { "backslash",	DEFAULT,	0x5C,	0x005C },
  { "bar",		DEFAULT,	0x7C,	0x007C },
  { "bet",		AGL2,		0x7A,	0x05D1 },
  { "beta",		AGL1,		0xC3,	0x03B2 },
  { "block",		AGL3,		0x8B,	0x2588 },
  { "braceex",		AGL4,		0x4C,	0xF8F4 },
  { "braceleft",	DEFAULT,	0x7B,	0x007B },
  { "braceleftbt",	AGL4,		0x4B,	0xF8F3 },
  { "braceleftmid",	AGL4,		0x4A,	0xF8F2 },
  { "bracelefttp",	AGL4,		0x49,	0xF8F1 },
  { "braceright",	DEFAULT,	0x7D,	0x007D },
  { "bracerightbt",	AGL4,		0x56,	0xF8FE },
  { "bracerightmid",	AGL4,		0x55,	0xF8FD },
  { "bracerighttp",	AGL4,		0x54,	0xF8FC },
  { "bracketleft",	DEFAULT,	0x5B,	0x005B },
  { "bracketleftbt",	AGL4,		0x48,	0xF8F0 },
  { "bracketleftex",	AGL4,		0x47,	0xF8EF },
  { "bracketlefttp",	AGL4,		0x46,	0xF8EE },
  { "bracketright",	DEFAULT,	0x5D,	0x005D },
  { "bracketrightbt",	AGL4,		0x53,	0xF8FB },
  { "bracketrightex",	AGL4,		0x52,	0xF8FA },
  { "bracketrighttp",	AGL4,		0x51,	0xF8F9 },
  { "breve",		AGL1,		0x8E,	0x02D8 },
  { "brokenbar",	DEFAULT,	0xA6,	0x00A6 },
  { "bsuperior",	AGL3,		0xD7,	0xF6EA },
  { "bullet",		DEFAULT,	0x7F,	0x2022 },
  { "bullet",		DEFAULT,	0x81,	0x2022 },
  { "bullet",		DEFAULT,	0x8D,	0x2022 },
  { "bullet",		DEFAULT,	0x8F,	0x2022 },
  { "bullet",		DEFAULT,	0x90,	0x2022 },
  { "bullet",		DEFAULT,	0x95,	0x2022 },
  { "bullet",		DEFAULT,	0x9D,	0x2022 },
  { "c",		DEFAULT,	0x63,	0x0063 },
  { "cacute",		AGL1,		0x08,	0x0107 },
  { "caron",		AGL1,		0x8C,	0x02C7 },
  { "carriagereturn",	AGL3,		0x2B,	0x21B5 },
  { "ccaron",		AGL1,		0x0E,	0x010D },
  { "ccedilla",		DEFAULT,	0xE7,	0x00E7 },
  { "ccircumflex",	AGL1,		0x0A,	0x0109 },
  { "cdotaccent",	AGL1,		0x0C,	0x010B },
  { "cedilla",		DEFAULT,	0xB8,	0x00B8 },
  { "cent",		DEFAULT,	0xA2,	0x00A2 },
  { "centinferior",	AGL3,		0xCC,	0xF6DF },
  { "centoldstyle",	AGL4,		0x18,	0xF7A2 },
  { "centsuperior",	AGL3,		0xCD,	0xF6E0 },
  { "chi",		AGL1,		0xD8,	0x03C7 },
  { "circle",		AGL3,		0x9B,	0x25CB },
  { "circlemultiply",	AGL3,		0x58,	0x2297 },
  { "circleplus",	AGL3,		0x57,	0x2295 },
  { "circumflex",	DEFAULT,	0x88,	0x02C6 },
  { "club",		AGL3,		0xA6,	0x2663 },
  { "colon",		DEFAULT,	0x3A,	0x003A },
  { "colonmonetary",	AGL3,		0x0E,	0x20A1 },
  { "comma",		DEFAULT,	0x2C,	0x002C },
  { "commaaccent",	AGL3,		0xB0,	0xF6C3 },
  { "commainferior",	AGL3,		0xCE,	0xF6E1 },
  { "commasuperior",	AGL3,		0xCF,	0xF6E2 },
  { "congruent",	AGL3,		0x4C,	0x2245 },
  { "copyright",	DEFAULT,	0xA9,	0x00A9 },
  { "copyrightsans",	AGL4,		0x41,	0xF8E9 },
  { "copyrightserif",	AGL3,		0xC6,	0xF6D9 },
  { "currency",		DEFAULT,	0xA4,	0x00A4 },
  { "cyrBreve",		AGL3,		0xBE,	0xF6D1 },
  { "cyrFlex",		AGL3,		0xBF,	0xF6D2 },
  { "cyrbreve",		AGL3,		0xC1,	0xF6D4 },
  { "cyrflex",		AGL3,		0xC2,	0xF6D5 },
  { "d",		DEFAULT,	0x64,	0x0064 },
  { "dagger",		DEFAULT,	0x86,	0x2020 },
  { "daggerdbl",	DEFAULT,	0x87,	0x2021 },
  { "dalet",		AGL2,		0x7C,	0x05D3 },
  { "dblGrave",		AGL3,		0xC0,	0xF6D3 },
  { "dblgrave",		AGL3,		0xC3,	0xF6D6 },
  { "dcaron",		AGL1,		0x10,	0x010F },
  { "dcroat",		AGL1,		0x12,	0x0111 },
  { "degree",		DEFAULT,	0xB0,	0x00B0 },
  { "delta",		AGL1,		0xC5,	0x03B4 },
  { "diamond",		AGL3,		0xA8,	0x2666 },
  { "dieresis",		DEFAULT,	0xA8,	0x00A8 },
  { "dieresisacute",	AGL3,		0xC4,	0xF6D7 },
  { "dieresisgrave",	AGL3,		0xC5,	0xF6D8 },
  { "dieresistonos",	AGL1,		0x99,	0x0385 },
  { "divide",		DEFAULT,	0xF7,	0x00F7 },
  { "dkshade",		AGL3,		0x90,	0x2593 },
  { "dnblock",		AGL3,		0x8A,	0x2584 },
  { "dollar",		DEFAULT,	0x24,	0x0024 },
  { "dollarinferior",	AGL3,		0xD0,	0xF6E3 },
  { "dollaroldstyle",	AGL3,		0xEE,	0xF724 },
  { "dollarsuperior",	AGL3,		0xD1,	0xF6E4 },
  { "dong",		AGL3,		0x13,	0x20AB },
  { "dotaccent",	AGL1,		0x8F,	0x02D9 },
  { "dotbelowcomb",	AGL1,		0x97,	0x0323 },
  { "dotlessi",		AGL1,		0x32,	0x0131 },
  { "dotlessj",		AGL3,		0xAB,	0xF6BE },
  { "dotmath",		AGL3,		0x5A,	0x22C5 },
  { "dsuperior",	AGL3,		0xD8,	0xF6EB },
  { "e",		DEFAULT,	0x65,	0x0065 },
  { "eacute",		DEFAULT,	0xE9,	0x00E9 },
  { "ebreve",		AGL1,		0x16,	0x0115 },
  { "ecaron",		AGL1,		0x1C,	0x011B },
  { "ecircumflex",	DEFAULT,	0xEA,	0x00EA },
  { "edieresis",	DEFAULT,	0xEB,	0x00EB },
  { "edotaccent",	AGL1,		0x18,	0x0117 },
  { "egrave",		DEFAULT,	0xE8,	0x00E8 },
  { "eight",		DEFAULT,	0x38,	0x0038 },
  { "eightinferior",	AGL3,		0x0A,	0x2088 },
  { "eightoldstyle",	AGL3,		0xF8,	0xF738 },
  { "eightsuperior",	AGL4,		0xFC,	0x2078 },
  { "element",		AGL3,		0x37,	0x2208 },
  { "ellipsis",		DEFAULT,	0x85,	0x2026 },
  { "emacron",		AGL1,		0x14,	0x0113 },
  { "emdash",		DEFAULT,	0x97,	0x2014 },
  { "emptyset",		AGL3,		0x34,	0x2205 },
  { "endash",		DEFAULT,	0x96,	0x2013 },
  { "eng",		AGL1,		0x4C,	0x014B },
  { "eogonek",		AGL1,		0x1A,	0x0119 },
  { "epsilon",		AGL1,		0xC6,	0x03B5 },
  { "epsilontonos",	AGL1,		0xBE,	0x03AD },
  { "equal",		DEFAULT,	0x3D,	0x003D },
  { "equivalence",	AGL3,		0x4F,	0x2261 },
  { "estimated",	AGL3,		0x1C,	0x212E },
  { "esuperior",	AGL3,		0xD9,	0xF6EC },
  { "eta",		AGL1,		0xC8,	0x03B7 },
  { "etatonos",		AGL1,		0xBF,	0x03AE },
  { "eth",		DEFAULT,	0xF0,	0x00F0 },
  { "exclam",		DEFAULT,	0x21,	0x0021 },
  { "exclamdbl",	AGL2,		0xF5,	0x203C },
  { "exclamdown",	DEFAULT,	0xA1,	0x00A1 },
  { "exclamdownsmall",	AGL4,		0x17,	0xF7A1 },
  { "exclamsmall",	AGL3,		0xED,	0xF721 },
  { "existential",	AGL3,		0x33,	0x2203 },
  { "f",		DEFAULT,	0x66,	0x0066 },
  { "female",		AGL3,		0xA3,	0x2640 },
  { "ff",		AGL4,		0x57,	0xFB00 },
  { "ffi",		AGL4,		0x5A,	0xFB03 },
  { "ffl",		AGL4,		0x5B,	0xFB04 },
  { "fi",		AGL4,		0x58,	0xFB01 },
  { "figuredash",	AGL2,		0xEA,	0x2012 },
  { "filledbox",	AGL3,		0x91,	0x25A0 },
  { "filledrect",	AGL3,		0x95,	0x25AC },
  { "finalkaf",		AGL2,		0x83,	0x05DA },
  { "finalmem",		AGL2,		0x86,	0x05DD },
  { "finalnun",		AGL2,		0x88,	0x05DF },
  { "finalpe",		AGL2,		0x8C,	0x05E3 },
  { "finaltsadi",	AGL2,		0x8E,	0x05E5 },
  { "five",		DEFAULT,	0x35,	0x0035 },
  { "fiveeighths",	AGL3,		0x22,	0x215D },
  { "fiveinferior",	AGL3,		0x07,	0x2085 },
  { "fiveoldstyle",	AGL3,		0xF5,	0xF735 },
  { "fivesuperior",	AGL4,		0xF9,	0x2075 },
  { "fl",		AGL4,		0x59,	0xFB02 },
  { "florin",		DEFAULT,	0x83,	0x0192 },
  { "four",		DEFAULT,	0x34,	0x0034 },
  { "fourinferior",	AGL3,		0x06,	0x2084 },
  { "fouroldstyle",	AGL3,		0xF4,	0xF734 },
  { "foursuperior",	AGL4,		0xF8,	0x2074 },
  { "fraction",		AGL2,		0xF6,	0x2044 },
  { "fraction",		AGL3,		0x3D,	0x2215 },
  { "franc",		AGL3,		0x0F,	0x20A3 },
  { "g",		DEFAULT,	0x67,	0x0067 },
  { "gamma",		AGL1,		0xC4,	0x03B3 },
  { "gbreve",		AGL1,		0x20,	0x011F },
  { "gcaron",		AGL1,		0x7F,	0x01E7 },
  { "gcircumflex",	AGL1,		0x1E,	0x011D },
  { "gcommaaccent",	AGL1,		0x24,	0x0123 },
  { "gdotaccent",	AGL1,		0x22,	0x0121 },
  { "germandbls",	DEFAULT,	0xDF,	0x00DF },
  { "gimel",		AGL2,		0x7B,	0x05D2 },
  { "gradient",		AGL3,		0x36,	0x2207 },
  { "grave",		DEFAULT,	0x60,	0x0060 },
  { "gravecomb",	AGL1,		0x93,	0x0300 },
  { "greater",		DEFAULT,	0x3E,	0x003E },
  { "greaterequal",	AGL3,		0x51,	0x2265 },
  { "guillemotleft",	DEFAULT,	0xAB,	0x00AB },
  { "guillemotright",	DEFAULT,	0xBB,	0x00BB },
  { "guilsinglleft",	DEFAULT,	0x8B,	0x2039 },
  { "guilsinglright",	DEFAULT,	0x9B,	0x203A },
  { "h",		DEFAULT,	0x68,	0x0068 },
  { "hbar",		AGL1,		0x28,	0x0127 },
  { "hcircumflex",	AGL1,		0x26,	0x0125 },
  { "he",		AGL2,		0x7D,	0x05D4 },
  { "heart",		AGL3,		0xA7,	0x2665 },
  { "het",		AGL2,		0x80,	0x05D7 },
  { "hookabovecomb",	AGL1,		0x96,	0x0309 },
  { "house",		AGL3,		0x5B,	0x2302 },
  { "hungarumlaut",	AGL1,		0x92,	0x02DD },
  { "hyphen",		DEFAULT,	0x2D,	0x002D },
  { "hyphen",		DEFAULT,	0xAD,	0x00AD },
  { "hypheninferior",	AGL3,		0xD2,	0xF6E5 },
  { "hyphensuperior",	AGL3,		0xD3,	0xF6E6 },
  { "i",		DEFAULT,	0x69,	0x0069 },
  { "iacute",		DEFAULT,	0xED,	0x00ED },
  { "ibreve",		AGL1,		0x2E,	0x012D },
  { "icircumflex",	DEFAULT,	0xEE,	0x00EE },
  { "idieresis",	DEFAULT,	0xEF,	0x00EF },
  { "igrave",		DEFAULT,	0xEC,	0x00EC },
  { "ij",		AGL1,		0x34,	0x0133 },
  { "imacron",		AGL1,		0x2C,	0x012B },
  { "infinity",		AGL3,		0x42,	0x221E },
  { "integral",		AGL3,		0x49,	0x222B },
  { "integralbt",	AGL3,		0x5E,	0x2321 },
  { "integralex",	AGL4,		0x4D,	0xF8F5 },
  { "integraltp",	AGL3,		0x5D,	0x2320 },
  { "intersection",	AGL3,		0x47,	0x2229 },
  { "invbullet",	AGL3,		0x9D,	0x25D8 },
  { "invcircle",	AGL3,		0x9E,	0x25D9 },
  { "invsmileface",	AGL3,		0xA1,	0x263B },
  { "iogonek",		AGL1,		0x30,	0x012F },
  { "iota",		AGL1,		0xCA,	0x03B9 },
  { "iotadieresis",	AGL1,		0xDB,	0x03CA },
  { "iotadieresistonos",	AGL1,		0xA2,	0x0390 },
  { "iotatonos",	AGL1,		0xC0,	0x03AF },
  { "isuperior",	AGL3,		0xDA,	0xF6ED },
  { "itilde",		AGL1,		0x2A,	0x0129 },
  { "j",		DEFAULT,	0x6A,	0x006A },
  { "jcircumflex",	AGL1,		0x36,	0x0135 },
  { "k",		DEFAULT,	0x6B,	0x006B },
  { "kaf",		AGL2,		0x84,	0x05DB },
  { "kappa",		AGL1,		0xCB,	0x03BA },
  { "kcommaaccent",	AGL1,		0x38,	0x0137 },
  { "kgreenlandic",	AGL1,		0x39,	0x0138 },
  { "l",		DEFAULT,	0x6C,	0x006C },
  { "lacute",		AGL1,		0x3B,	0x013A },
  { "lambda",		AGL1,		0xCC,	0x03BB },
  { "lamed",		AGL2,		0x85,	0x05DC },
  { "lcaron",		AGL1,		0x3F,	0x013E },
  { "lcommaaccent",	AGL1,		0x3D,	0x013C },
  { "ldot",		AGL1,		0x41,	0x0140 },
  { "less",		DEFAULT,	0x3C,	0x003C },
  { "lessequal",	AGL3,		0x50,	0x2264 },
  { "lfblock",		AGL3,		0x8C,	0x258C },
  { "lira",		AGL3,		0x10,	0x20A4 },
  { "ll",		AGL3,		0xAD,	0xF6C0 },
  { "logicaland",	AGL3,		0x45,	0x2227 },
  { "logicalnot",	DEFAULT,	0xAC,	0x00AC },
  { "logicalor",	AGL3,		0x46,	0x2228 },
  { "longs",		AGL1,		0x79,	0x017F },
  { "lozenge",		AGL3,		0x9A,	0x25CA },
  { "lslash",		AGL1,		0x43,	0x0142 },
  { "lsuperior",	AGL3,		0xDB,	0xF6EE },
  { "ltshade",		AGL3,		0x8E,	0x2591 },
  { "m",		DEFAULT,	0x6D,	0x006D },
  { "macron",		AGL1,		0x8D,	0x02C9 },
  { "macron",		DEFAULT,	0xAF,	0x00AF },
  { "male",		AGL3,		0xA4,	0x2642 },
  { "mem",		AGL2,		0x87,	0x05DE },
  { "minus",		AGL3,		0x3C,	0x2212 },
  { "minute",		AGL2,		0xF3,	0x2032 },
  { "msuperior",	AGL3,		0xDC,	0xF6EF },
  { "mu",		AGL1,		0xCD,	0x03BC },
  { "mu",		DEFAULT,	0xB5,	0x00B5 },
  { "multiply",		DEFAULT,	0xD7,	0x00D7 },
  { "musicalnote",	AGL3,		0xA9,	0x266A },
  { "musicalnotedbl",	AGL3,		0xAA,	0x266B },
  { "n",		DEFAULT,	0x6E,	0x006E },
  { "nacute",		AGL1,		0x45,	0x0144 },
  { "napostrophe",	AGL1,		0x4A,	0x0149 },
  { "nbspace",		DEFAULT,	0xA0,	0x00A0 },	/* 'nbspace' in glyphlist.txt 2.0 */
  { "ncaron",		AGL1,		0x49,	0x0148 },
  { "ncommaaccent",	AGL1,		0x47,	0x0146 },
  { "nine",		DEFAULT,	0x39,	0x0039 },
  { "nineinferior",	AGL3,		0x0B,	0x2089 },
  { "nineoldstyle",	AGL3,		0xF9,	0xF739 },
  { "ninesuperior",	AGL4,		0xFD,	0x2079 },
  { "notelement",	AGL3,		0x38,	0x2209 },
  { "notequal",		AGL3,		0x4E,	0x2260 },
  { "notsubset",	AGL3,		0x54,	0x2284 },
  { "nsuperior",	AGL4,		0xF6,	0x207F },
  { "ntilde",		DEFAULT,	0xF1,	0x00F1 },
  { "nu",		AGL1,		0xCE,	0x03BD },
  { "numbersign",	DEFAULT,	0x23,	0x0023 },
  { "nun",		AGL2,		0x89,	0x05E0 },
  { "o",		DEFAULT,	0x6F,	0x006F },
  { "oacute",		DEFAULT,	0xF3,	0x00F3 },
  { "obreve",		AGL1,		0x50,	0x014F },
  { "ocircumflex",	DEFAULT,	0xF4,	0x00F4 },
  { "odieresis",	DEFAULT,	0xF6,	0x00F6 },
  { "oe",		DEFAULT,	0x9C,	0x0153 },
  { "ogonek",		AGL1,		0x91,	0x02DB },
  { "ograve",		DEFAULT,	0xF2,	0x00F2 },
  { "ohorn",		AGL1,		0x7B,	0x01A1 },
  { "ohungarumlaut",	AGL1,		0x52,	0x0151 },
  { "omacron",		AGL1,		0x4E,	0x014D },
  { "omega",		AGL1,		0xDA,	0x03C9 },
  { "omega1",		AGL1,		0xE3,	0x03D6 },
  { "omegatonos",	AGL1,		0xDF,	0x03CE },
  { "omicron",		AGL1,		0xD0,	0x03BF },
  { "omicrontonos",	AGL1,		0xDD,	0x03CC },
  { "one",		DEFAULT,	0x31,	0x0031 },
  { "onedotenleader",	AGL2,		0xEE,	0x2024 },
  { "oneeighth",	AGL3,		0x20,	0x215B },
  { "onefitted",	AGL3,		0xC9,	0xF6DC },
  { "onehalf",		DEFAULT,	0xBD,	0x00BD },
  { "oneinferior",	AGL3,		0x03,	0x2081 },
  { "oneoldstyle",	AGL3,		0xF1,	0xF731 },
  { "onequarter",	DEFAULT,	0xBC,	0x00BC },
  { "onesuperior",	DEFAULT,	0xB9,	0x00B9 },
  { "onethird",		AGL3,		0x1E,	0x2153 },
  { "openbullet",	AGL3,		0x9F,	0x25E6 },
  { "ordfeminine",	DEFAULT,	0xAA,	0x00AA },
  { "ordmasculine",	DEFAULT,	0xBA,	0x00BA },
  { "orthogonal",	AGL3,		0x43,	0x221F },
  { "oslash",		DEFAULT,	0xF8,	0x00F8 },
  { "oslashacute",	AGL1,		0x85,	0x01FF },
  { "osuperior",	AGL3,		0xDD,	0xF6F0 },
  { "otilde",		DEFAULT,	0xF5,	0x00F5 },
  { "p",		DEFAULT,	0x70,	0x0070 },
  { "paragraph",	DEFAULT,	0xB6,	0x00B6 },
  { "parenleft",	DEFAULT,	0x28,	0x0028 },
  { "parenleftbt",	AGL4,		0x45,	0xF8ED },
  { "parenleftex",	AGL4,		0x44,	0xF8EC },
  { "parenleftinferior",	AGL3,		0x0C,	0x208D },
  { "parenleftsuperior",	AGL4,		0xFE,	0x207D },
  { "parenlefttp",	AGL4,		0x43,	0xF8EB },
  { "parenright",	DEFAULT,	0x29,	0x0029 },
  { "parenrightbt",	AGL4,		0x50,	0xF8F8 },
  { "parenrightex",	AGL4,		0x4F,	0xF8F7 },
  { "parenrightinferior",	AGL3,		0x0D,	0x208E },
  { "parenrightsuperior",	AGL4,		0xFF,	0x207E },
  { "parenrighttp",	AGL4,		0x4E,	0xF8F6 },
  { "partialdiff",	AGL3,		0x32,	0x2202 },
  { "pe",		AGL2,		0x8D,	0x05E4 },
  { "percent",		DEFAULT,	0x25,	0x0025 },
  { "period",		DEFAULT,	0x2E,	0x002E },
  { "periodcentered",	AGL3,		0x3F,	0x2219 },
  { "periodcentered",	DEFAULT,	0xB7,	0x00B7 },
  { "periodinferior",	AGL3,		0xD4,	0xF6E7 },
  { "periodsuperior",	AGL3,		0xD5,	0xF6E8 },
  { "perpendicular",	AGL3,		0x59,	0x22A5 },
  { "perthousand",	DEFAULT,	0x89,	0x2030 },
  { "peseta",		AGL3,		0x11,	0x20A7 },
  { "phi",		AGL1,		0xD7,	0x03C6 },
  { "phi1",		AGL1,		0xE2,	0x03D5 },
  { "pi",		AGL1,		0xD1,	0x03C0 },
  { "plus",		DEFAULT,	0x2B,	0x002B },
  { "plusminus",	DEFAULT,	0xB1,	0x00B1 },
  { "prescription",	AGL3,		0x1A,	0x211E },
  { "product",		AGL3,		0x3A,	0x220F },
  { "propersubset",	AGL3,		0x52,	0x2282 },
  { "propersuperset",	AGL3,		0x53,	0x2283 },
  { "proportional",	AGL3,		0x41,	0x221D },
  { "psi",		AGL1,		0xD9,	0x03C8 },
  { "q",		DEFAULT,	0x71,	0x0071 },
  { "qof",		AGL2,		0x90,	0x05E7 },
  { "question",		DEFAULT,	0x3F,	0x003F },
  { "questiondown",	DEFAULT,	0xBF,	0x00BF },
  { "questiondownsmall",	AGL4,		0x1D,	0xF7BF },
  { "questionsmall",	AGL3,		0xFA,	0xF73F },
  { "quotedbl",		DEFAULT,	0x22,	0x0022 },
  { "quotedblbase",	DEFAULT,	0x84,	0x201E },
  { "quotedblleft",	DEFAULT,	0x93,	0x201C },
  { "quotedblright",	DEFAULT,	0x94,	0x201D },
  { "quoteleft",	DEFAULT,	0x91,	0x2018 },
  { "quotereversed",	AGL2,		0xED,	0x201B },
  { "quoteright",	DEFAULT,	0x92,	0x2019 },
  { "quotesinglbase",	DEFAULT,	0x82,	0x201A },
  { "quotesingle",	DEFAULT,	0x27,	0x0027 },
  { "r",		DEFAULT,	0x72,	0x0072 },
  { "racute",		AGL1,		0x54,	0x0155 },
  { "radical",		AGL3,		0x40,	0x221A },
  { "radicalex",	AGL4,		0x3D,	0xF8E5 },
  { "rcaron",		AGL1,		0x58,	0x0159 },
  { "rcommaaccent",	AGL1,		0x56,	0x0157 },
  { "reflexsubset",	AGL3,		0x55,	0x2286 },
  { "reflexsuperset",	AGL3,		0x56,	0x2287 },
  { "registered",	DEFAULT,	0xAE,	0x00AE },
  { "registersans",	AGL4,		0x40,	0xF8E8 },
  { "registerserif",	AGL3,		0xC7,	0xF6DA },
  { "resh",		AGL2,		0x91,	0x05E8 },
  { "revlogicalnot",	AGL3,		0x5C,	0x2310 },
  { "rho",		AGL1,		0xD2,	0x03C1 },
  { "ring",		AGL1,		0x90,	0x02DA },
  { "rsuperior",	AGL3,		0xDE,	0xF6F1 },
  { "rtblock",		AGL3,		0x8D,	0x2590 },
  { "rupiah",		AGL3,		0xCA,	0xF6DD },
  { "s",		DEFAULT,	0x73,	0x0073 },
  { "sacute",		AGL1,		0x5A,	0x015B },
  { "samekh",		AGL2,		0x8A,	0x05E1 },
  { "scaron",		DEFAULT,	0x9A,	0x0161 },
  { "scedilla",		AGL1,		0x5E,	0x015F },
  { "scircumflex",	AGL1,		0x5C,	0x015D },
  { "scommaaccent",	AGL1,		0x87,	0x0219 },
  { "second",		AGL2,		0xF4,	0x2033 },
  { "section",		DEFAULT,	0xA7,	0x00A7 },
  { "semicolon",	DEFAULT,	0x3B,	0x003B },
  { "seven",		DEFAULT,	0x37,	0x0037 },
  { "seveneighths",	AGL3,		0x23,	0x215E },
  { "seveninferior",	AGL3,		0x09,	0x2087 },
  { "sevenoldstyle",	AGL3,		0xF7,	0xF737 },
  { "sevensuperior",	AGL4,		0xFB,	0x2077 },
  { "shade",		AGL3,		0x8F,	0x2592 },
  { "shin",		AGL2,		0x92,	0x05E9 },
  { "sigma",		AGL1,		0xD4,	0x03C3 },
  { "sigma1",		AGL1,		0xD3,	0x03C2 },
  { "similar",		AGL3,		0x4B,	0x223C },
  { "six",		DEFAULT,	0x36,	0x0036 },
  { "sixinferior",	AGL3,		0x08,	0x2086 },
  { "sixoldstyle",	AGL3,		0xF6,	0xF736 },
  { "sixsuperior",	AGL4,		0xFA,	0x2076 },
  { "slash",		DEFAULT,	0x2F,	0x002F },
  { "smileface",	AGL3,		0xA0,	0x263A },
  { "space",		DEFAULT,	0x20,	0x0020 },
  { "spade",		AGL3,		0xA5,	0x2660 },
  { "ssuperior",	AGL3,		0xDF,	0xF6F2 },
  { "sterling",		DEFAULT,	0xA3,	0x00A3 },
  { "suchthat",		AGL3,		0x39,	0x220B },
  { "summation",	AGL3,		0x3B,	0x2211 },
  { "sun",		AGL3,		0xA2,	0x263C },
  { "t",		DEFAULT,	0x74,	0x0074 },
  { "tau",		AGL1,		0xD5,	0x03C4 },
  { "tav",		AGL2,		0x93,	0x05EA },
  { "tbar",		AGL1,		0x64,	0x0167 },
  { "tcaron",		AGL1,		0x62,	0x0165 },
  { "tcommaaccent",	AGL1,		0x60,	0x0163 },
  { "tet",		AGL2,		0x81,	0x05D8 },
  { "therefore",	AGL3,		0x4A,	0x2234 },
  { "theta",		AGL1,		0xC9,	0x03B8 },
  { "theta1",		AGL1,		0xE0,	0x03D1 },
  { "thorn",		DEFAULT,	0xFE,	0x00FE },
  { "three",		DEFAULT,	0x33,	0x0033 },
  { "threeeighths",	AGL3,		0x21,	0x215C },
  { "threeinferior",	AGL3,		0x05,	0x2083 },
  { "threeoldstyle",	AGL3,		0xF3,	0xF733 },
  { "threequarters",	DEFAULT,	0xBE,	0x00BE },
  { "threequartersemdash",	AGL3,		0xCB,	0xF6DE },
  { "threesuperior",	DEFAULT,	0xB3,	0x00B3 },
  { "tilde",		DEFAULT,	0x98,	0x02DC },
  { "tildecomb",	AGL1,		0x95,	0x0303 },
  { "tonos",		AGL1,		0x98,	0x0384 },
  { "trademark",	DEFAULT,	0x99,	0x2122 },
  { "trademarksans",	AGL4,		0x42,	0xF8EA },
  { "trademarkserif",	AGL3,		0xC8,	0xF6DB },
  { "triagdn",		AGL3,		0x98,	0x25BC },
  { "triaglf",		AGL3,		0x99,	0x25C4 },
  { "triagrt",		AGL3,		0x97,	0x25BA },
  { "triagup",		AGL3,		0x96,	0x25B2 },
  { "tsadi",		AGL2,		0x8F,	0x05E6 },
  { "tsuperior",	AGL3,		0xE0,	0xF6F3 },
  { "two",		DEFAULT,	0x32,	0x0032 },
  { "twodotenleader",	AGL2,		0xEF,	0x2025 },
  { "twoinferior",	AGL3,		0x04,	0x2082 },
  { "twooldstyle",	AGL3,		0xF2,	0xF732 },
  { "twosuperior",	DEFAULT,	0xB2,	0x00B2 },
  { "twothirds",	AGL3,		0x1F,	0x2154 },
  { "u",		DEFAULT,	0x75,	0x0075 },
  { "uacute",		DEFAULT,	0xFA,	0x00FA },
  { "ubreve",		AGL1,		0x6A,	0x016D },
  { "ucircumflex",	DEFAULT,	0xFB,	0x00FB },
  { "udieresis",	DEFAULT,	0xFC,	0x00FC },
  { "ugrave",		DEFAULT,	0xF9,	0x00F9 },
  { "uhorn",		AGL1,		0x7D,	0x01B0 },
  { "uhungarumlaut",	AGL1,		0x6E,	0x0171 },
  { "umacron",		AGL1,		0x68,	0x016B },
  { "underscore",	DEFAULT,	0x5F,	0x005F },
  { "underscoredbl",	AGL2,		0xEC,	0x2017 },
  { "union",		AGL3,		0x48,	0x222A },
  { "universal",	AGL3,		0x31,	0x2200 },
  { "uogonek",		AGL1,		0x70,	0x0173 },
  { "upblock",		AGL3,		0x89,	0x2580 },
  { "upsilon",		AGL1,		0xD6,	0x03C5 },
  { "upsilondieresis",	AGL1,		0xDC,	0x03CB },
  { "upsilondieresistonos",	AGL1,		0xC1,	0x03B0 },
  { "upsilontonos",	AGL1,		0xDE,	0x03CD },
  { "uring",		AGL1,		0x6C,	0x016F },
  { "utilde",		AGL1,		0x66,	0x0169 },
  { "v",		DEFAULT,	0x76,	0x0076 },
  { "vav",		AGL2,		0x7E,	0x05D5 },
  { "w",		DEFAULT,	0x77,	0x0077 },
  { "wacute",		AGL2,		0xE1,	0x1E83 },
  { "wcircumflex",	AGL1,		0x72,	0x0175 },
  { "wdieresis",	AGL2,		0xE3,	0x1E85 },
  { "weierstrass",	AGL3,		0x18,	0x2118 },
  { "wgrave",		AGL2,		0xDF,	0x1E81 },
  { "x",		DEFAULT,	0x78,	0x0078 },
  { "xi",		AGL1,		0xCF,	0x03BE },
  { "y",		DEFAULT,	0x79,	0x0079 },
  { "yacute",		DEFAULT,	0xFD,	0x00FD },
  { "ycircumflex",	AGL1,		0x74,	0x0177 },
  { "ydieresis",	DEFAULT,	0xFF,	0x00FF },
  { "yen",		DEFAULT,	0xA5,	0x00A5 },
  { "ygrave",		AGL2,		0xE5,	0x1EF3 },
  { "yod",		AGL2,		0x82,	0x05D9 },
  { "z",		DEFAULT,	0x7A,	0x007A },
  { "zacute",		AGL1,		0x76,	0x017A },
  { "zayin",		AGL2,		0x7F,	0x05D6 },
  { "zcaron",		DEFAULT,	0x9E,	0x017E },
  { "zdotaccent",	AGL1,		0x78,	0x017C },
  { "zero",		DEFAULT,	0x30,	0x0030 },
  { "zeroinferior",	AGL3,		0x02,	0x2080 },
  { "zerooldstyle",	AGL3,		0xF0,	0xF730 },
  { "zerosuperior",	AGL4,		0xF7,	0x2070 },
  { "zeta",		AGL1,		0xC7,	0x03B6 },
};

PangoPDFLIBFont *
_pango_pdflib_font_new (PangoFontMap    *fontmap,
			FcPattern  *pattern)
{
  PangoPDFLIBFont *pdflibfont;
  double d;

  g_return_val_if_fail (fontmap != NULL, NULL);
  g_return_val_if_fail (pattern != NULL, NULL);

  pdflibfont = (PangoPDFLIBFont *)g_object_new (PANGO_TYPE_PDFLIB_FONT, NULL);

  pdflibfont->fontmap = fontmap;
  pdflibfont->font_pattern = pattern;
  pdflibfont->postscript_name = NULL;
  pdflibfont->font_handle = -1;
  pdflibfont->agl1_font_handle = -1;
  pdflibfont->agl2_font_handle = -1;
  pdflibfont->agl3_font_handle = -1;
  pdflibfont->agl4_font_handle = -1;

  g_object_ref (G_OBJECT (fontmap));
  pdflibfont->description = _pango_pdflib_font_desc_from_pattern (pattern, TRUE);
  pdflibfont->face = NULL;

  if (FcPatternGetDouble (pattern, FC_PIXEL_SIZE, 0, &d) == FcResultMatch)
    pdflibfont->size = d*PANGO_SCALE;

  _pango_pdflib_font_map_add (pdflibfont->fontmap, pdflibfont);

  return pdflibfont;
}

static gboolean
set_unicode_charmap (FT_Face face)
{
  int charmap;

  for (charmap = 0; charmap < face->num_charmaps; charmap++)
    if (face->charmaps[charmap]->encoding == ft_encoding_unicode)
      {
	FT_Error error = FT_Set_Charmap(face, face->charmaps[charmap]);
	return error == FT_Err_Ok;
      }

  return FALSE;
}

static void
load_fallback_face (PangoPDFLIBFont *pdflibfont,
		    const char   *original_file)
{
  FcPattern *sans;
  FcPattern *matched;
  FcResult result;
  FT_Error error;
  FcChar8 *filename2 = NULL;
  gchar *name;
  int id;
  
  sans = FcPatternBuild (NULL,
			 FC_FAMILY, FcTypeString, "sans",
			 FC_SIZE, FcTypeDouble, (double)pango_font_description_get_size (pdflibfont->description)/PANGO_SCALE,
			 NULL);
  
  matched = FcFontMatch (0, sans, &result);
  
  if (FcPatternGetString (matched, FC_FILE, 0, &filename2) != FcResultMatch)
    goto bail1;
  
  if (FcPatternGetInteger (matched, FC_INDEX, 0, &id) != FcResultMatch)
    goto bail1;
  
  error = FT_New_Face (_pango_pdflib_font_map_get_library (pdflibfont->fontmap),
		       (char *) filename2, id, &pdflibfont->face);
  
  
  if (error)
    {
    bail1:
      name = pango_font_description_to_string (pdflibfont->description);
      g_warning ("Unable to open font file %s for font %s, exiting\n", filename2, name);
      exit (1);
    }
  else
    {
      name = pango_font_description_to_string (pdflibfont->description);
      g_warning ("Unable to open font file %s for font %s, falling back to %s\n", original_file, name, filename2);
      g_free (name);
    }

  if (!set_unicode_charmap (pdflibfont->face))
    {
      g_warning ("Unable to load unicode charmap from file %s, exiting\n", filename2);
      exit (1);
    }
  
  FcPatternDestroy (sans);
  FcPatternDestroy (matched);
}

/**
 * pango_pdflib_font_get_face:
 * @font: a #PangoFont
 * 
 * Returns the native FreeType2 FT_Face structure used for this PangoFont.
 * This may be useful if you want to use FreeType2 functions directly.
 * 
 * Return value: a pointer to a #FT_Face structure, with the size set correctly
 **/
FT_Face
pango_pdflib_font_get_face (PangoFont      *font)
{
  PangoPDFLIBFont *pdflibfont = (PangoPDFLIBFont *)font;
  FT_Error error;
  FcPattern *pattern;
  FcChar8 *filename;
  FcBool antialias, hinting, autohint;
  int id;

  pattern = pdflibfont->font_pattern;

  if (!pdflibfont->face)
    {
      pdflibfont->load_flags = 0;

      /* disable antialiasing if requested */
      if (FcPatternGetBool (pattern,
                            FC_ANTIALIAS, 0, &antialias) != FcResultMatch)
	antialias = FcTrue;

      if (antialias)
        pdflibfont->load_flags |= FT_LOAD_NO_BITMAP;
      else
	pdflibfont->load_flags |= FT_LOAD_TARGET_MONO;

      /* disable hinting if requested */
      if (FcPatternGetBool (pattern,
                            FC_HINTING, 0, &hinting) != FcResultMatch)
	hinting = FcTrue;

      if (!hinting)
        pdflibfont->load_flags |= FT_LOAD_NO_HINTING;

      /* force autohinting if requested */
      if (FcPatternGetBool (pattern,
                            FC_AUTOHINT, 0, &autohint) != FcResultMatch)
	autohint = FcFalse;

      if (autohint)
        pdflibfont->load_flags |= FT_LOAD_FORCE_AUTOHINT;

      if (FcPatternGetString (pattern, FC_FILE, 0, &filename) != FcResultMatch)
	      goto bail0;
      
      if (FcPatternGetInteger (pattern, FC_INDEX, 0, &id) != FcResultMatch)
	      goto bail0;

      error = FT_New_Face (_pango_pdflib_font_map_get_library (pdflibfont->fontmap),
			   (char *) filename, id, &pdflibfont->face);
      if (error != FT_Err_Ok)
	{
	bail0:
	  load_fallback_face (pdflibfont, filename);
	}

      g_assert (pdflibfont->face);
  
      if (!set_unicode_charmap (pdflibfont->face))
	{
	  g_warning ("Unable to load unicode charmap from font file %s",
                     filename);
      
	  FT_Done_Face (pdflibfont->face);
	  pdflibfont->face = NULL;

	  load_fallback_face (pdflibfont, filename);
	}

      error = FT_Set_Char_Size (pdflibfont->face,
				PANGO_PIXELS_26_6 (pdflibfont->size),
				PANGO_PIXELS_26_6 (pdflibfont->size),
				72 * PANGO_SCALE / 64,
				72 * PANGO_SCALE / 64);

      /*
      if (error)
	g_warning ("Error in FT_Set_Char_Size: %d", error);
      */
    }
  
  return pdflibfont->face;
}

static GType
pango_pdflib_font_get_type (void)
{
  static GType object_type = 0;

  if (!object_type)
    {
      static const GTypeInfo object_info =
      {
        sizeof (PangoPDFLIBFontClass),
        (GBaseInitFunc) NULL,
        (GBaseFinalizeFunc) NULL,
        (GClassInitFunc) pango_pdflib_font_class_init,
        NULL,           /* class_finalize */
        NULL,           /* class_data */
        sizeof (PangoPDFLIBFont),
        0,              /* n_preallocs */
        (GInstanceInitFunc) pango_pdflib_font_init,
      };
      
      object_type = g_type_register_static (PANGO_TYPE_FONT,
                                            "PangoPDFLIBFont",
                                            &object_info, 0);
    }
  
  return object_type;
}

static void 
pango_pdflib_font_init (PangoPDFLIBFont *pdflibfont)
{
  pdflibfont->face = NULL;

  pdflibfont->size = 0;

  pdflibfont->metrics_by_lang = NULL;

  pdflibfont->glyph_info = g_hash_table_new (NULL, NULL);
}

static void
pango_pdflib_font_class_init (PangoPDFLIBFontClass *class)
{
  GObjectClass *object_class = G_OBJECT_CLASS (class);
  PangoFontClass *font_class = PANGO_FONT_CLASS (class);

  parent_class = g_type_class_peek_parent (class);
  
  object_class->finalize = pango_pdflib_font_finalize;
  
  font_class->describe = pango_pdflib_font_describe;
  font_class->get_coverage = pango_pdflib_font_get_coverage;
  font_class->find_shaper = pango_pdflib_font_find_shaper;
  font_class->get_glyph_extents = pango_pdflib_font_get_glyph_extents;
  font_class->get_metrics = pango_pdflib_font_get_metrics;
}

static void
maybe_set_afm (PDF *pdf_doc,
	       const gchar *fontname,
	       const gchar *non_afm_filename)
{
  GString *filename = g_string_new (non_afm_filename);
  int len = strlen (non_afm_filename);

  g_string_truncate (filename,
		     len-3);
  g_string_append (filename,
		   "afm");
  if (g_file_test (filename->str, G_FILE_TEST_EXISTS))
    {
      GString *parameter = g_string_new (NULL);

      g_string_printf (parameter,
		       "%s==%s",
		       fontname,
		       filename->str);

      PDF_set_parameter (pdf_doc,
			 "FontAFM",
			 parameter->str);

      g_string_free (parameter, TRUE);
    }
  g_string_free (filename, TRUE);
}

#if 0
static void
pango_pdflib_debug_dump_glyph (PangoFont *font,
			       PangoGlyphInfo *gi,
			       gchar     *this_char)
{
  PangoRectangle ink, logical;
  FT_Face face;

  pango_font_get_glyph_extents (font,
				gi->glyph,
				&ink,
				&logical);

  face = pango_pdflib_font_get_face (font);

  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "      glyph ink: x='%1.1f' y='%1.1f' w='%1.1f' h='%1.1f'",
	 (float) ink.x / PANGO_SCALE,
	 (float) ink.y / PANGO_SCALE,
	 (float) ink.width / PANGO_SCALE,
	 (float) ink.height / PANGO_SCALE);
  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "            log: x='%1.1f' y='%1.1f' w='%1.1f' h='%1.1f'",
	 (float) logical.x / PANGO_SCALE,
	 (float) logical.y / PANGO_SCALE,
	 (float) logical.width / PANGO_SCALE,
	 (float) logical.height / PANGO_SCALE);
  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "            off: x='%1.1f' y='%1.1f'",
	 (float) gi->geometry.x_offset,
	 (float) gi->geometry.y_offset);
  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "            geo_w='%1.1f' fname='%s' sname='%s'",
	 (float) gi->geometry.width,
	 face->family_name,
	 face->style_name);
  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "            units_per_EM='%d' ascender='%d' descender='%d' height='%d'",
	 face->units_per_EM,
	 face->ascender,
	 face->descender,
	 face->height);
  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "            font='%s'",
	 pango_font_description_to_string (pango_font_describe (font)));
  g_log (G_LOG_DOMAIN,
	 G_LOG_LEVEL_DEBUG,
	 "            char='%04x' kern='%s'",
	 *this_char,
	 FT_HAS_KERNING(face) ? "yes" : "no");
}
#endif

typedef struct _FontInfo {
  const gchar *postscript_name;
  const gchar *use_postscript_name;
  const gchar *encoding_name;
  gboolean     embed;
} FontInfo;

static FontInfo font_entries[] = {
  {"Courier",                 NULL,                   NULL,            TRUE},
  {"Courier-Bold",            NULL,                   NULL,            TRUE},
  {"Courier-BoldOblique",     NULL,                   NULL,            TRUE},
  {"Courier-Oblique",         NULL,                   NULL,            TRUE},
  {"HYSMyeongJo-Medium",      "HYSMyeongJo-Medium",   "UniKS-UCS2-H",  FALSE},
  {"HYSMyeongJo-Medium-Acro", "HYSMyeongJo-Medium",   "UniKS-UCS2-H",  FALSE},
  {"HeiseiKakuGo-W5",         "HeiseiKakuGo-W5",      "UniJIS-UCS2-H", FALSE},
  {"HeiseiKakuGo-W5-Acro",    "HeiseiKakuGo-W5",      "UniJIS-UCS2-H", FALSE},
  {"HeiseiMin-W3",            "HeiseiMin-W3",         "UniJIS-UCS2-H", FALSE},
  {"HeiseiMin-W3-Acro",       "HeiseiMin-W3",         "UniJIS-UCS2-H", FALSE},
  {"Helvetica",               NULL,                   NULL,            TRUE},
  {"Helvetica-Bold",          NULL,                   NULL,            TRUE},
  {"Helvetica-BoldOblique",   NULL,                   NULL,            TRUE},
  {"Helvetica-Oblique",       NULL,                   NULL,            TRUE},
  {"STSong-Light",            "STSong-Light",         "UniGB-UCS2-H",  FALSE},
  {"STSong-Light-Acro",       "STSong-Light",         "UniGB-UCS2-H",  FALSE},
  {"Symbol",                  NULL,                   NULL,            TRUE},
  {"Times-Bold",              NULL,                   NULL,            TRUE},
  {"Times-BoldItalic",        NULL,                   NULL,            TRUE},
  {"Times-Italic",            NULL,                   NULL,            TRUE},
  {"Times-Roman",             NULL,                   NULL,            TRUE},
  {"ZapfDingbats",            NULL,                   NULL,            TRUE}
};

static int
pango_pdflib_compare_font_info (const void *a, const void *b)
{
  return strcmp ((const char *) a, ((const FontInfo *) b)->postscript_name);
}

static int
pango_pdflib_compare_glyphs (const void *a, const void *b)
{
  return strcmp ((const char *) a, ((const GlyphEntry *) b)->name);
}

static void
pango_pdflib_load_default_font (PDF       *pdf_doc,
				PangoFont *font)
{
  PangoPDFLIBFont *pdflibfont = (PangoPDFLIBFont *) font;
  FcChar8 *family = NULL;
  FcChar8 *style = NULL;
  FcChar8 *font_filename = NULL;
  FcPattern *pattern = NULL;
  FcResult fc_result;
  FontInfo *font_info = NULL;
  const gchar *use_encoding;

  pattern = pdflibfont->font_pattern;

  fc_result = FcPatternGetString (pattern, FC_FILE, 0, &font_filename);

  fc_result = FcPatternGetString (pattern, FC_FAMILY, 0, &family);
  fc_result = FcPatternGetString (pattern, FC_STYLE, 0, &style);

  font_info =
    (FontInfo *) bsearch (pdflibfont->postscript_name,
			  font_entries,
			  G_N_ELEMENTS (font_entries),
			  sizeof (FontInfo),
			  pango_pdflib_compare_font_info);

  if (font_info != NULL && font_info->encoding_name != NULL)
    {
      use_encoding = font_info->encoding_name;
    }
  else
    {
      FcLangSet *lang_set;

      fc_result = FcPatternGetLangSet (pattern, FC_LANG, 0, &lang_set);
      if (fc_result == FcResultMatch)
	{
	  if (FcLangSetHasLang (lang_set, "ja") == FcLangEqual)
	    {
	      use_encoding = "UniJIS-UCS2-H";
	    }
	  else if (FcLangSetHasLang (lang_set, "ko") == FcLangEqual)
	    {
	      use_encoding = "UniKS-UCS2-H";
	    }
	  else
	    {
	      FT_CharMapRec *charmap =
		((PangoPDFLIBFont *) font)->face->charmap;

	      /* PDFlib expects "winansi" when platform_id == 3 and
		 encoding_id == 1 */
	      if ((charmap->platform_id == 3) &&
		  (charmap->encoding_id == 1))
		{
		  use_encoding = "winansi";
		}
	      else
		{
		  use_encoding = "builtin";
		}
	    }
	}
      else
	{
	  FT_CharMapRec *charmap =
	    ((PangoPDFLIBFont *) font)->face->charmap;

	  if ((charmap->platform_id == 3) &&
	      (charmap->encoding_id == 1))
	    {
	      use_encoding = "winansi";
	    }
	  else
	    {
	      use_encoding = "builtin";
	    }
	}
    }

  /* The default font handle */
  ((PangoPDFLIBFont *) font)->font_handle =
    PDF_findfont (pdf_doc,
		  pdflibfont->postscript_name,
		  use_encoding,
		  1);
}

static void
pango_pdflib_enable_font (PDF       *pdf_doc,
			  PangoFont *font)
{
  PangoPDFLIBFont *pdflibfont = (PangoPDFLIBFont *) font;
  FcChar8 *family = NULL;
  FcChar8 *style = NULL;
  FcChar8 *font_filename = NULL;
  FcPattern *pattern = NULL;
  FcResult fc_result;
  FontInfo *font_info = NULL;
  GString *parameter = g_string_new (NULL);
  int len;

  pattern = pdflibfont->font_pattern;

  fc_result = FcPatternGetString (pattern, FC_FILE, 0, &font_filename);

  fc_result = FcPatternGetString (pattern, FC_FAMILY, 0, &family);
  fc_result = FcPatternGetString (pattern, FC_STYLE, 0, &style);

  if (pdflibfont->postscript_name == NULL)
    {
      pdflibfont->postscript_name =
	g_strdup (FT_Get_Postscript_Name (pango_pdflib_font_get_face (font)));
    }
  else
    {
      g_assert_not_reached ();
    }

  font_info =
    (FontInfo *) bsearch (pdflibfont->postscript_name,
			  font_entries,
			  G_N_ELEMENTS (font_entries),
			  sizeof (FontInfo),
			  pango_pdflib_compare_font_info);

  if (font_info != NULL && font_info->postscript_name != NULL)
    {
      g_free (pdflibfont->postscript_name);
      pdflibfont->postscript_name = g_strdup (font_info->use_postscript_name);
    }
  else
    {
      FcLangSet *lang_set;
      int postscript_len = strlen (pdflibfont->postscript_name);

      if (g_ascii_strncasecmp (&pdflibfont->postscript_name[postscript_len - 5],
			       "-Acro",
			       5) == 0)
	{
	  gchar *tmp_name;

	  tmp_name = pdflibfont->postscript_name;

	  pdflibfont->postscript_name =
	    g_strndup (tmp_name,
		       postscript_len - 5);
	}

      fc_result = FcPatternGetLangSet (pattern, FC_LANG, 0, &lang_set);
      if (fc_result == FcResultMatch)
	{
	  if (FcLangSetHasLang (lang_set, "ja") == FcLangEqual)
	    {
	      g_free (pdflibfont->postscript_name);
	      pdflibfont->postscript_name = g_strdup ("HeiseiMin-W3");
	    }
	  else if (FcLangSetHasLang (lang_set, "ko") == FcLangEqual)
	    {
	      g_free (pdflibfont->postscript_name);
	      pdflibfont->postscript_name = g_strdup ("HYSMyeongJo-Medium");
	    }
	}
    }


  g_string_printf (parameter,
		   "%s==%s",
		   pdflibfont->postscript_name,
		   font_filename);

  len = strlen (font_filename);

  if (len > 4)
    {
      if (g_ascii_strncasecmp (&font_filename[len-4], ".ttf", 4) == 0)
	{
	  PDF_set_parameter (pdf_doc,
			     "FontOutline",
			     parameter->str);
	}
      if ((g_ascii_strncasecmp (&font_filename[len-4], ".pfa", 4) == 0) ||
	  (g_ascii_strncasecmp (&font_filename[len-4], ".pfb", 4) == 0))
	{
	  PDF_set_parameter (pdf_doc,
			     "FontOutline",
			     parameter->str);
	  maybe_set_afm (pdf_doc,
			 pdflibfont->postscript_name,
			 font_filename);
	}
      else if (g_ascii_strncasecmp (&font_filename[len-4], ".pfm", 4) == 0)
	{
	  PDF_set_parameter (pdf_doc,
			     "FontPFM",
			     parameter->str);
	  maybe_set_afm (pdf_doc,
			 pdflibfont->postscript_name,
			 font_filename);
	}
      else if (g_ascii_strncasecmp (&font_filename[len-4], ".afm", 4) == 0)
	{
	  PDF_set_parameter (pdf_doc,
			     "FontPFM",
			     parameter->str);
	}
    }

  g_string_free (parameter, TRUE);
}

/**
 * pango_pdflib_render:
 * @pdf_doc:     the PDFlib document onto which to draw @layout_text.
 * @layout_text: the text of the layout.
 * @run:         the #PangoLayoutRun to render.
 * @font:        the font in which to draw the string.
 * @glyphs:      the glyph string to draw.
 * @x:           the x position of the start of the string (in #PangoGlyphUnit).
 * @y:           the y position of the baseline (in #PangoGlyphUnit).
 *
 * Renders a PangoGlyphString onto a PDFlib PDF document.
 **/
void
pango_pdflib_render (PDF              *pdf_doc,
		     const gchar      *layout_text,
		     PangoLayoutRun   *run,
		     PangoFont        *font,
		     PangoGlyphString *glyphs,
		     gint              x, 
		     gint              y)
{
  /*PangoLayoutRunPrivate *prun = (PangoLayoutRunPrivate *) run;*/
  /*FT_UInt glyph_index;*/
  int i;
  int x_position = 0;
  PangoPDFLIBFont *pdflibfont = ((PangoPDFLIBFont *) font);
  PangoGlyphInfo *gi;
  GSList *extra_attrs_list = run->item->analysis.extra_attrs;
  gboolean foreground_set = FALSE;
  guint16 foreground_red = 0, foreground_green = 0, foreground_blue = 0;
  gboolean background_set = FALSE;
  guint16 background_red = 0, background_green = 0, background_blue = 0;
  gint rise = 0;
  gint baseline_adjust = 0;
  gboolean font_is_set = FALSE;
  FcPattern *pattern;
  FcResult fc_result;
  FcChar8 *font_filename;
  double font_size;

  g_return_if_fail (pdf_doc != NULL);
  g_return_if_fail (run != NULL);
  g_return_if_fail (glyphs != NULL);

  while (extra_attrs_list)
    {
      PangoAttribute *attr = extra_attrs_list->data;
      PangoAttrType attr_type = attr->klass->type;

      switch (attr_type)
	{
	case PANGO_ATTR_RISE:
	  rise = ((PangoAttrInt *) attr)->value;
	  break;
	case PANGO_ATTR_FOREGROUND:
	  foreground_set = TRUE;
	  foreground_red = ((PangoAttrColor *) attr)->color.red;
	  foreground_green = ((PangoAttrColor *) attr)->color.green;
	  foreground_blue = ((PangoAttrColor *) attr)->color.blue;
	  break;
	case PANGO_ATTR_BACKGROUND:
	  background_set = TRUE;
	  background_red = ((PangoAttrColor *) attr)->color.red;
	  background_green = ((PangoAttrColor *) attr)->color.green;
	  background_blue = ((PangoAttrColor *) attr)->color.blue;
	  break;
	default:
	  if (attr_type == pango_attr_callback_get_type ())
	    {
	      GValue values[1] = { { 0, } };

	      g_value_init (&values[0], G_TYPE_POINTER);
	      g_value_set_pointer (&values[0], pdf_doc);

#if 0
	      g_message ("Got a callback: callback: %p",
			 ((PangoAttrPointer *) attr)->pointer);
#endif

	      PDF_translate (pdf_doc,
			     ((float) x) / PANGO_SCALE,
			     ((float) y) / PANGO_SCALE);

	      g_closure_invoke (((GClosure *) ((PangoAttrPointer *) attr)->pointer),
				NULL,
				1,
				values,
				NULL);
	      PDF_translate (pdf_doc,
			     - (((float) x) / PANGO_SCALE),
			     - (((float) y) / PANGO_SCALE));
	    }
	  break;
	}
      extra_attrs_list = extra_attrs_list->next;
    }

  if (background_set)
    {
      PangoRectangle logical;

      pango_glyph_string_extents (glyphs,
				  font,
				  NULL,
				  &logical);
      PDF_setcolor (pdf_doc,
		    "fill",
		    "rgb",
		    (float) background_red / 0xFFFF,
		    (float) background_green / 0xFFFF,
		    (float) background_blue / 0xFFFF,
		    0);

      PDF_rect (pdf_doc,
		((float) (x + logical.x)) / PANGO_SCALE,
		((float) y + rise + baseline_adjust - logical.y) / PANGO_SCALE,
		((float) logical.width) / PANGO_SCALE,
		((float) -logical.height) / PANGO_SCALE);
      PDF_fill (pdf_doc);
    }

  if (foreground_set)
    {
      PDF_setcolor (pdf_doc,
		    "fill",
		    "rgb",
		    (float) foreground_red / 0xFFFF,
		    (float) foreground_green / 0xFFFF,
		    (float) foreground_blue / 0xFFFF,
		    0);
    }

  pattern = pdflibfont->font_pattern;

  fc_result = FcPatternGetString (pattern, FC_FILE, 0, &font_filename);
  if (fc_result == FcResultMatch)
    {

      fc_result = FcPatternGetDouble (pattern, FC_SIZE, 0, &font_size);
      
      if (pdflibfont->postscript_name == NULL)
	{
	  pango_pdflib_enable_font (pdf_doc,
				  font);
	}
    }
  else
    {
      g_warning ("No font file for font.");
      return;
    }

  gi = glyphs->glyphs;
  for (i = 0; i < glyphs->num_glyphs; i++, gi++)
    {
      if (gi->glyph)
	{
	  gunichar unichar;
	  gunichar2 *utf16_char;
	  gint current_handle = -1;

	  unichar =
	    g_utf8_get_char(&layout_text[run->item->offset +
					glyphs->log_clusters[i]]);


#if 0
	  g_utf8_strncpy (this_char,
			  &layout_text[run->item->offset + glyphs->log_clusters[i]],
			  1);

	  pango_pdflib_debug_dump_glyph (font,
					 gi,
					 this_char);
#endif

	  if (g_unichar_iswide (unichar))
	    {
	      glong utf16_char_count;
	      gint i;

	      if (pdflibfont->font_handle == -1)
		{
		  pango_pdflib_load_default_font (pdf_doc,
						  font);
		}

	      if (font_is_set == TRUE)
		{
		  current_handle =
		    (int) PDF_get_value (pdf_doc, "font", 0);
		}
	      else
		{
		  current_handle = -1;
		}

	      if (current_handle != pdflibfont->font_handle)
		{
		  PDF_setfont (pdf_doc,
			       pdflibfont->font_handle,
			       font_size);
		  font_is_set = TRUE;
		}

#if 0
	      g_log (G_LOG_DOMAIN,
		     G_LOG_LEVEL_DEBUG,
		     "Glyph:: U+%4X",
		     unichar);
#endif
	      utf16_char =
		g_ucs4_to_utf16 (&unichar,
				 1,
				 NULL,
				 &utf16_char_count,
				 NULL);

	      /* x86 is little-endian, but PDF expects big-endian */
	      for (i = 0; i < (int) utf16_char_count; i++)
		{
		  utf16_char[i] = GUINT16_TO_BE (utf16_char[i]);
		}

	      PDF_show_xy2 (pdf_doc,
			    (gchar *) utf16_char,
			    ((int) utf16_char_count) * 2,
			    ((float) (x + x_position)) / PANGO_SCALE,
			    ((float) (y + rise + baseline_adjust)) / PANGO_SCALE);

	      g_free (utf16_char);
	    }
	  else
	    {
	      gchar glyph_name[100];
	      guchar code_point = 0;
	      GlyphEntry *glyph_entry;

	      FT_Get_Glyph_Name(pdflibfont->face,
				gi->glyph & 0xFFFF,
				glyph_name,
				sizeof glyph_name);
	      glyph_entry =
		(GlyphEntry *) bsearch (glyph_name,
					glyph_entries,
					G_N_ELEMENTS (glyph_entries),
					sizeof (GlyphEntry),
					pango_pdflib_compare_glyphs);

#if 0
	      g_log (G_LOG_DOMAIN,
		     G_LOG_LEVEL_DEBUG,
		     "Glyph:: %s",
		     glyph_name);
#endif

	      if (glyph_entry != NULL)
		{
		  gint correct_handle = -1;
		  gint current_handle = -1;

		  code_point = glyph_entry->code_point;

		  switch (glyph_entry->encoding)
		    {
		    case AGL1:
		      correct_handle = pdflibfont->agl1_font_handle;

		      if (correct_handle == -1)
			{
			  PDF_set_parameter (pdf_doc,
					     "Encoding",
					     "agl1==" PKGDATADIR "/" MODULE_VERSION "/agl1.enc");
			  correct_handle =
			    pdflibfont->agl1_font_handle =
			    PDF_findfont (pdf_doc,
					  pdflibfont->postscript_name,
					  "agl1",
					  1);
			}
		      break;
		    case AGL2:
		      correct_handle = pdflibfont->agl2_font_handle;

		      if (correct_handle == -1)
			{
			  PDF_set_parameter (pdf_doc,
					     "Encoding",
					     "agl2==" PKGDATADIR "/" MODULE_VERSION "/agl2.enc");
			  correct_handle =
			    pdflibfont->agl2_font_handle =
			    PDF_findfont (pdf_doc,
					  pdflibfont->postscript_name,
					  "agl2",
					  1);
			}
		      break;
		    case AGL3:
		      correct_handle = pdflibfont->agl3_font_handle;

		      if (correct_handle == -1)
			{
			  PDF_set_parameter (pdf_doc,
					     "Encoding",
					     "agl3==" PKGDATADIR "/" MODULE_VERSION "/agl3.enc");
			  correct_handle =
			    pdflibfont->agl3_font_handle =
			    PDF_findfont (pdf_doc,
					  pdflibfont->postscript_name,
					  "agl3",
					  1);
			}
		      break;
		    case AGL4:
		      correct_handle = pdflibfont->agl4_font_handle;

		      if (correct_handle == -1)
			{
			  PDF_set_parameter (pdf_doc,
					     "Encoding",
					     "agl4==" PKGDATADIR "/" MODULE_VERSION "/agl4.enc");
			  correct_handle =
			    pdflibfont->agl4_font_handle =
			    PDF_findfont (pdf_doc,
					  pdflibfont->postscript_name,
					  "agl4",
					  1);
			}
		      break;
		    default:
		      correct_handle = pdflibfont->font_handle;

		      if (correct_handle == -1)
			{
			  pango_pdflib_load_default_font (pdf_doc,
							  font);

			  correct_handle = pdflibfont->font_handle;
			}
		      break;
		    }

		  if (font_is_set == TRUE)
		    {
		      current_handle =
			(int) PDF_get_value (pdf_doc, "font", 0);
		    }
		  else
		    {
		      current_handle = -1;
		    }

#if 0
		  g_log (G_LOG_DOMAIN,
			 G_LOG_LEVEL_DEBUG,
			 "        enc: %d; cp: %02X; unicode: U+%04X",
			 glyph_entry->encoding,
			 glyph_entry->code_point,
			 glyph_entry->unicode);
		  g_log (G_LOG_DOMAIN,
			 G_LOG_LEVEL_DEBUG,
			 "        handle: %d; correct handle: %d",
			 current_handle,
			 correct_handle);
#endif

		  if (current_handle != correct_handle)
		    {
		      PDF_setfont (pdf_doc,
				   correct_handle,
				   font_size);
		      font_is_set = TRUE;
		    }
		}

	      PDF_show_xy2 (pdf_doc,
			    &code_point,
			    1,
			    ((float) (x + x_position)) / PANGO_SCALE,
			    ((float) (y + rise + baseline_adjust)) / PANGO_SCALE);
	    }
	}

      x_position += glyphs->glyphs[i].geometry.width;
    }
}

static FT_Glyph_Metrics *
pango_pdflib_get_per_char (PangoFont *font,
			guint32    glyph_index)
{
  PangoPDFLIBFont *pdflibfont = (PangoPDFLIBFont *)font;
  FT_Face face;
	
  face = pango_pdflib_font_get_face (font);
  
  FT_Load_Glyph (face, glyph_index, pdflibfont->load_flags);
  return &face->glyph->metrics;
}

static PangoPDFLIBGlyphInfo *
pango_pdflib_font_get_glyph_info (PangoFont   *font,
			       PangoGlyph   glyph,
			       gboolean     create)
{
  PangoPDFLIBFont *pdflibfont = (PangoPDFLIBFont *)font;
  PangoPDFLIBGlyphInfo *info;
  FT_Glyph_Metrics *gm;

  info = g_hash_table_lookup (pdflibfont->glyph_info, GUINT_TO_POINTER (glyph));

  if ((info == NULL) && create)
    {
      FT_Face face = pango_pdflib_font_get_face (font);
      info = g_new0 (PangoPDFLIBGlyphInfo, 1);
      
      if (glyph && (gm = pango_pdflib_get_per_char (font, glyph)))
	{
#if 0
	  info->ink_rect.x = PANGO_UNITS_26_6_PX (gm->horiBearingX);
	  info->ink_rect.width = PANGO_UNITS_26_6_PX (gm->width);
	  info->ink_rect.y = -PANGO_UNITS_26_6_PX (gm->horiBearingY);
	  info->ink_rect.height = PANGO_UNITS_26_6_PX (gm->height);
	      
	  info->logical_rect.x = 0;
	  info->logical_rect.width = PANGO_UNITS_26_6_PX (gm->horiAdvance);
	  info->logical_rect.y = -PANGO_UNITS_26_6_PX (face->size->metrics.ascender + 64);
	  /* Some fonts report negative descender, some positive ! (?) */
	  info->logical_rect.height = PANGO_UNITS_26_6_PX (face->size->metrics.ascender + ABS (face->size->metrics.descender) + 128);
#endif
	  info->ink_rect.x = gm->horiBearingX;
	  info->ink_rect.width = gm->width;
	  info->ink_rect.y = -gm->horiBearingY;
	  info->ink_rect.height = gm->height;
	      
	  info->logical_rect.x = 0;
	  info->logical_rect.width = gm->horiAdvance;
	  info->logical_rect.y = -face->size->metrics.ascender;
	  /* Some fonts report negative descender, some positive ! (?) */
	  info->logical_rect.height = face->size->metrics.ascender + ABS (face->size->metrics.descender);
	}
      else
	{
	  info->ink_rect.x = 0;
	  info->ink_rect.width = 0;
	  info->ink_rect.y = 0;
	  info->ink_rect.height = 0;

	  info->logical_rect.x = 0;
	  info->logical_rect.width = 0;
	  info->logical_rect.y = 0;
	  info->logical_rect.height = 0;
	}

      g_hash_table_insert (pdflibfont->glyph_info, GUINT_TO_POINTER(glyph), info);
    }

  return info;
}
     
static void
pango_pdflib_font_get_glyph_extents (PangoFont      *font,
				  PangoGlyph      glyph,
				  PangoRectangle *ink_rect,
				  PangoRectangle *logical_rect)
{
  PangoPDFLIBGlyphInfo *info;

  info = pango_pdflib_font_get_glyph_info (font, glyph, TRUE);

  if (ink_rect)
    *ink_rect = info->ink_rect;
  if (logical_rect)
    *logical_rect = info->logical_rect;
}

/**
 * pango_pdflib_font_get_kerning:
 * @font: a #PangoFont
 * @left: the left #PangoGlyph
 * @right: the right #PangoGlyph
 * 
 * Retrieves kerning information for a combination of two glyphs.
 * 
 * Return value: The amount of kerning (in Pango units) to apply for 
 * the given combination of glyphs.
 **/
int
pango_pdflib_font_get_kerning (PangoFont *font,
			    PangoGlyph left,
			    PangoGlyph right)
{
  FT_Face face;
  FT_Error error;
  FT_Vector kerning;

  face = pango_pdflib_font_get_face (font);
  if (!face)
    return 0;

  if (!FT_HAS_KERNING (face))
    return 0;

  if (!left || !right)
    return 0;

  error = FT_Get_Kerning (face, left, right,
			  ft_kerning_default, &kerning);
  if (error != FT_Err_Ok)
    g_warning ("FT_Get_Kerning returns error: %s",
	       _pango_pdflib_ft_strerror (error));

#if 0
  return PANGO_UNITS_26_6_PX (kerning.x);
#endif
  return kerning.x;
}

static PangoFontMetrics *
pango_pdflib_font_get_metrics (PangoFont     *font,
			    PangoLanguage *language)
{
  PangoPDFLIBFont *pdflibfont = PANGO_PDFLIB_FONT (font);
  PangoPDFLIBMetricsInfo *info = NULL; /* Quiet gcc */
  GSList *tmp_list;      

  const char *sample_str = pango_language_get_sample_string (language);
  
  tmp_list = pdflibfont->metrics_by_lang;
  while (tmp_list)
    {
      info = tmp_list->data;
      
      if (info->sample_str == sample_str)    /* We _don't_ need strcmp */
	break;

      tmp_list = tmp_list->next;
    }

  if (!tmp_list)
    {
      PangoContext *context;
      PangoLayout *layout;
      PangoRectangle extents;
      FT_Face face = pango_pdflib_font_get_face (font);

      info = g_new (PangoPDFLIBMetricsInfo, 1);
      info->sample_str = sample_str;
      info->metrics = pango_font_metrics_new ();

#if 0
      info->metrics->ascent = PANGO_UNITS_26_6 (face->size->metrics.ascender);
      info->metrics->descent = PANGO_UNITS_26_6 (- face->size->metrics.descender);
      info->metrics->approximate_char_width = 
        info->metrics->approximate_digit_width = 
        PANGO_UNITS_26_6 (face->size->metrics.max_advance);
#endif
      info->metrics->ascent = face->size->metrics.ascender;
      info->metrics->descent = - face->size->metrics.descender;
      info->metrics->approximate_char_width = 
        info->metrics->approximate_digit_width = 
        face->size->metrics.max_advance;

      pdflibfont->metrics_by_lang = g_slist_prepend (pdflibfont->metrics_by_lang, info);

      context = pango_context_new ();
      pango_context_set_font_map (context, pdflibfont->fontmap);
      pango_context_set_language (context, language);
      
      layout = pango_layout_new (context);
      pango_layout_set_font_description (layout, pdflibfont->description);

      pango_layout_set_text (layout, sample_str, -1);      
      pango_layout_get_extents (layout, NULL, &extents);
      
      info->metrics->approximate_char_width = 
        extents.width / g_utf8_strlen (sample_str, -1);

      pango_layout_set_text (layout, "0123456789", -1);      
      pango_layout_get_extents (layout, NULL, &extents);
      
      info->metrics->approximate_digit_width = extents.width / 10;

      g_object_unref (G_OBJECT (layout));
      g_object_unref (G_OBJECT (context));
    }

  return pango_font_metrics_ref (info->metrics);
}

static gboolean
pango_pdflib_free_glyph_info_callback (gpointer key, gpointer value, gpointer data)
{
  PangoPDFLIBFont *font = PANGO_PDFLIB_FONT (data);
  PangoPDFLIBGlyphInfo *info = value;
  
  if (font->glyph_cache_destroy && info->cached_glyph)
    (*font->glyph_cache_destroy) (info->cached_glyph);

  g_free (value);
  return TRUE;
}

static void
free_metrics_info (PangoPDFLIBMetricsInfo *info)
{
  pango_font_metrics_unref (info->metrics);
  g_free (info);
}

static void
pango_pdflib_font_finalize (GObject *object)
{
  PangoPDFLIBFont *pdflibfont = (PangoPDFLIBFont *)object;

  _pango_pdflib_font_map_remove (pdflibfont->fontmap, pdflibfont);

  if (pdflibfont->face)
    {
      FT_Done_Face (pdflibfont->face);
      pdflibfont->face = NULL;
    }

  pango_font_description_free (pdflibfont->description);
  FcPatternDestroy (pdflibfont->font_pattern);
  
  g_object_unref (G_OBJECT (pdflibfont->fontmap));

  g_free (pdflibfont->postscript_name);

  g_slist_foreach (pdflibfont->metrics_by_lang, (GFunc)free_metrics_info, NULL);
  g_slist_free (pdflibfont->metrics_by_lang);  

  g_hash_table_foreach_remove (pdflibfont->glyph_info,
			       pango_pdflib_free_glyph_info_callback, object);
  g_hash_table_destroy (pdflibfont->glyph_info);
  
  G_OBJECT_CLASS (parent_class)->finalize (object);
}

static PangoFontDescription *
pango_pdflib_font_describe (PangoFont *font)
{
  PangoPDFLIBFont         *pdflibfont;
  PangoFontDescription *desc;

  pdflibfont = PANGO_PDFLIB_FONT (font);

  desc = pango_font_description_copy (pdflibfont->description);

  return desc;
}

PangoMap *
pango_pdflib_get_shaper_map (PangoLanguage *language)
{
  static guint engine_type_id = 0;
  static guint render_type_id = 0;
  
  if (engine_type_id == 0)
    {
      engine_type_id = g_quark_from_static_string (PANGO_ENGINE_TYPE_SHAPE);
      render_type_id = g_quark_from_static_string (PANGO_RENDER_TYPE_PDFLIB);
    }
  
  return pango_find_map (language, engine_type_id, render_type_id);
}

/**
 * pango_pdflib_font_get_coverage:
 * @font: a #PangoPDFLIBFont.
 * @language: a language tag.
 * @returns: a #PangoCoverage.
 * 
 * Should not be called directly, use pango_font_get_coverage() instead.
 **/
PangoCoverage *
pango_pdflib_font_get_coverage (PangoFont     *font,
			     PangoLanguage *language)
{
  PangoPDFLIBFont *pdflibfont = (PangoPDFLIBFont *)font;

  return _pango_pdflib_font_map_get_coverage (pdflibfont->fontmap, pdflibfont->font_pattern);
}

static PangoEngineShape *
pango_pdflib_font_find_shaper (PangoFont     *font,
			    PangoLanguage *language,
			    guint32        ch)
{
  PangoMap *shape_map = NULL;

  shape_map = pango_pdflib_get_shaper_map (language);
  return (PangoEngineShape *)pango_map_get_engine (shape_map, ch);
}

/* Utility functions */

/**
 * pango_pdflib_get_unknown_glyph:
 * @font: a #PangoFont
 * 
 * Return the index of a glyph suitable for drawing unknown characters.
 * 
 * Return value: a glyph index into @font
 **/
PangoGlyph
pango_pdflib_get_unknown_glyph (PangoFont *font)
{
  return 0;
}


/**
 * pango_pdflib_render_layout_line:
 * @pdf_doc:     the PDFlib document onto which to draw @layout_text.
 * @layout_text: the text of the layout.
 * @line:        a #PangoLayoutLine
 * @x:           the x position of start of string (in #PangoGlyphUnit)
 * @y:           the y position of baseline (in #PangoGlyphUnit)
 *
 * Render a #PangoLayoutLine onto a FreeType2 bitmap
 */
void 
pango_pdflib_render_layout_line (PDF             *pdf_doc,
				 const gchar     *layout_text,
				 PangoLayoutLine *line,
				 gint             x, 
				 gint             y)
{
  GSList *run_list;
  PangoRectangle overall_rect;
  PangoRectangle logical_rect;
  gint      x_off = 0;

  pango_layout_line_get_extents (line,
				 NULL,
				 &overall_rect);
  
  run_list = line->runs;
  while (run_list)
    {
      PangoLayoutRun *run = run_list->data;
      
      pango_glyph_string_extents (run->glyphs,
				  run->item->analysis.font,
				  NULL,
				  &logical_rect);

      pango_pdflib_render (pdf_doc,
			   layout_text,
			   run,
			   run->item->analysis.font,
			   run->glyphs,
			   x + x_off,
			   y);

      x_off += logical_rect.width;
      run_list = run_list->next;
    }
}

/**
 * pango_pdflib_render_layout_lines:
 * @pdf_doc:     the PDFlib document onto which to draw @layout_text.
 * @layout:      a #PangoLayout
 * @line_first:  the first line to render.
 * @line_last:   the last line to render.
 * @x:           the X position of the left of the layout (in #PangoGlyphUnit)
 * @y:           the Y position of the top of the layout (in #PangoGlyphUnit)
 *
 * Render some lines from #PangoLayout onto a PDFlib PDF document.
 */
void 
pango_pdflib_render_layout_lines (PDF         *pdf_doc,
				  PangoLayout *layout,
				  gint	       line_first,
				  gint         line_last,
				  gint         x, 
				  gint         y)
{
  PangoLayoutIter *iter;
  const gchar *layout_text = pango_layout_get_text (layout);
  
  g_return_if_fail (pdf_doc != NULL);
  g_return_if_fail (PANGO_IS_LAYOUT (layout));
  g_return_if_fail (line_first >= 0);
  g_return_if_fail (line_last >= line_first &&
		    line_last <= g_slist_length (pango_layout_get_lines (layout)) - 1);

  iter = pango_layout_get_iter (layout);

  gint line_number = -1;
  do
    {
      PangoRectangle   logical_rect;
      PangoLayoutLine *line;
      int              baseline;
      
      line_number++;

      if (line_number < line_first)
	{
	  continue;
	}

      line = pango_layout_iter_get_line (iter);
      
      pango_layout_iter_get_line_extents (iter, NULL, &logical_rect);
      baseline = pango_layout_iter_get_baseline (iter);
      
      pango_pdflib_render_layout_line (pdf_doc,
				       layout_text,
				       line,
				       x + logical_rect.x,
				       y - baseline);

      if (line_number >= line_last)
	{
	  break;
	}
    }
  while (pango_layout_iter_next_line (iter));

  pango_layout_iter_free (iter);
}

/**
 * pango_pdflib_render_layout:
 * @pdf_doc:     the PDFlib document onto which to render @layout.
 * @layout:      a #PangoLayout.
 * @x:           the X position of the left of the layout (in #PangoGlyphUnit).
 * @y:           the Y position of the top of the layout (in #PangoGlyphUnit).
 *
 * Render a #PangoLayout onto a PDFlib PDF document.
 **/
void 
pango_pdflib_render_layout (PDF         *pdf_doc,
			    PangoLayout *layout,
			    gint         x, 
			    gint         y)
{
  gint line_last;

  g_return_if_fail (pdf_doc != NULL);
  g_return_if_fail (PANGO_IS_LAYOUT (layout));

  line_last =
    ((gint) g_slist_length (pango_layout_get_lines (layout))) - 1;

  pango_pdflib_render_layout_lines (pdf_doc,
				    layout,
				    0,
				    line_last,
				    x,
				    y);
}

#if 0
/* This utility function is duplicated here and in pango-layout.c; should it be
 * public? Trouble is - what is the appropriate set of properties?
 */
static void
pango_pdflib_get_item_properties (PangoItem      *item,
				  PangoUnderline *uline,
				  PangoAttrColor *fg_color,
				  gboolean       *fg_set,
				  PangoAttrColor *bg_color,
				  gboolean       *bg_set)
{
  GSList *tmp_list = item->analysis.extra_attrs;

  if (fg_set)
    *fg_set = FALSE;
  
  if (bg_set)
    *bg_set = FALSE;
  
  while (tmp_list)
    {
      PangoAttribute *attr = tmp_list->data;

      switch (attr->klass->type)
	{
	case PANGO_ATTR_UNDERLINE:
	  if (uline)
	    *uline = ((PangoAttrInt *)attr)->value;
	  break;
	  
	case PANGO_ATTR_FOREGROUND:
	  if (fg_color)
	    *fg_color = *((PangoAttrColor *)attr);
	  if (fg_set)
	    *fg_set = TRUE;
	  
	  break;
	  
	case PANGO_ATTR_BACKGROUND:
	  if (bg_color)
	    *bg_color = *((PangoAttrColor *)attr);
	  if (bg_set)
	    *bg_set = TRUE;
	  
	  break;
	  
	default:
	  break;
	}
      tmp_list = tmp_list->next;
    }
}
#endif

typedef struct
{
  FT_Error     code;
  const char*  msg;
} ft_error_description;

static int
ft_error_compare (const void *pkey,
		  const void *pbase)
{
  return ((ft_error_description *) pkey)->code - ((ft_error_description *) pbase)->code;
}

const char *
_pango_pdflib_ft_strerror (FT_Error error)
{
#undef __FTERRORS_H__
#define FT_ERRORDEF( e, v, s )  { e, s },
#define FT_ERROR_START_LIST  {
#define FT_ERROR_END_LIST    { 0, 0 } };

  const ft_error_description ft_errors[] =
#include FT_ERRORS_H

#undef FT_ERRORDEF
#undef FT_ERROR_START_LIST
#undef FT_ERROR_END_LIST

  ft_error_description *found =
    bsearch (&error, ft_errors, G_N_ELEMENTS (ft_errors) - 1,
	     sizeof (ft_errors[0]), ft_error_compare);
  if (found != NULL)
    return found->msg;
  else
    {
      static char default_msg[100];

      sprintf (default_msg, "Unknown FreeType2 error %#x", error);
      return default_msg;
    }
}


void *
pango_pdflib_font_get_cache_glyph_data (PangoFont *font,
				     int        glyph_index)
{
  PangoPDFLIBGlyphInfo *info;

  g_return_val_if_fail (PANGO_PDFLIB_IS_FONT (font), NULL);
  
  info = pango_pdflib_font_get_glyph_info (font, glyph_index, FALSE);

  if (info == NULL)
    return NULL;
  
  return info->cached_glyph;
}

void
pango_pdflib_font_set_cache_glyph_data (PangoFont     *font,
				     int            glyph_index,
				     void          *cached_glyph)
{
  PangoPDFLIBGlyphInfo *info;

  g_return_if_fail (PANGO_PDFLIB_IS_FONT (font));
  
  info = pango_pdflib_font_get_glyph_info (font, glyph_index, TRUE);

  info->cached_glyph = cached_glyph;

  /* TODO: Implement limiting of the number of cached glyphs */
}

void
pango_pdflib_font_set_glyph_cache_destroy (PangoFont      *font,
					GDestroyNotify  destroy_notify)
{
  g_return_if_fail (PANGO_PDFLIB_IS_FONT (font));
  
  PANGO_PDFLIB_FONT (font)->glyph_cache_destroy = destroy_notify;
}
